%if style == newcode
\begin{code}
  {-# OPTIONS --universe-polymorphism #-}
  module appendix where
    open import Level public
    open import Data.Sum public hiding (map)
    open import Relation.Binary.PropositionalEquality public hiding ([_] ; _≗_)

    Indexed : Set -> Set₁
    Indexed I = I -> Set

    _▷_ : Set -> Set -> Set₁
    I ▷ O = Indexed I -> Indexed O

    infix 3 _≃_
    record _≃_ (A B : Set) : Set where
      field
        from  : A -> B
        to    : B -> A
        iso₁  : ∀ x -> to (from x) ≡ x
        iso₂  : ∀ x -> from (to x) ≡ x

    _∣_ : {I J : Set} -> Indexed I -> Indexed J -> Indexed (I ⊎ J)
    (r ∣ s)  (inj₁  i)  = r  i
    (r ∣ s)  (inj₂  j)  = s  j

    _⇉_ : {I : Set} -> Indexed I -> Indexed I -> Set
    r ⇉ s = ∀ i -> r i -> s i

    -- Merging natural transformations.
    infixr 5 _∥_
    _∥_ : {I J : Set} {r u : Indexed I} {s v : Indexed J} ->
          (r ⇉ u) -> (s ⇉ v) -> ((r ∣ s) ⇉ (u ∣ v))
    _∥_ f g (inj₁ x) = f x
    _∥_ f g (inj₂ y) = g y

    data ∃ {A : Set} (B : A -> Set) : Set where
      some : ∀ {x} -> B x -> ∃ B 
\end{code}

\begin{spec}
    data ⊥ : Set where

    data ⊤ : Set where
      tt : ⊤

    infixr 6 _×_
    infixr 6 _,_

    data _×_ (A B : Set) : Set where
      _,_ : A -> B -> A × B

    infixr 5 _+_

    data _+_ (A B : Set) : Set where 
      inj₁ : A -> A + B
      inj₂ : B -> A + B

    const : ∀ {a b : Level} {A : Set a} {B : Set b} -> A -> B -> A
    const x _ = x

    uncurry : {A B C : Set} -> (A -> B -> C) -> (A × B) -> C
    uncurry f (x , y) = f x y

    id : {A : Set} -> A -> A
    id x = x

    infixr 7 _∘_ 
    _∘_ : ∀ {a b c}
            {A : Set a} {B : A -> Set b} {C : {x : A} -> B x -> Set c} ->
          (∀ {x} (y : B x) -> C y) -> (g : (x : A) -> B x) ->
          ((x : A) -> C (g x))
    (f ∘ g) x = f (g x)
\end{spec}


Type equality and utilities
\begin{spec}
    infix 7 _≡_
  
    data _≡_ {ℓ : Level} {A : Set ℓ} (x : A) : A -> Set ℓ where
      refl : x ≡ x
\end{spec}

\begin{spec}
    infix 3 _≃_
    record _≃_ (A B : Set) : Set where
      field
        from  : A -> B
        to    : B -> A
        iso₁  : ∀ {x} -> to (from x) ≡ x
        iso₂  : ∀ {x} -> from (to x) ≡ x

    Indexed : Set -> Set₁
    Indexed I = I -> Set

    _▷_ : Set -> Set -> Set₁
    I ▷ O = Indexed I -> Indexed O
{-
    T : Set -> Indexed ⊤
    T A tt = A
-}
    -- Merging indexed families.
    infixr 5 _∣_
    _∣_ : ∀ {I J} -> Indexed I -> Indexed J -> Indexed (I + J)
    (r ∣ s) (inj₁ i) = r i
    (r ∣ s) (inj₂ j) = s j

    -- Natural transformations
    _⇉_ : ∀ {I} -> Indexed I -> Indexed I -> Set
    r ⇉ s = ∀ i -> r i -> s i

    -- Composing natural transformations
    _⇉∘_ : {I : Set} {r s t : Indexed I} -> s ⇉ t -> r ⇉ s -> r ⇉ t
    (f ⇉∘ g) i x = f i (g i x)
\end{spec}


Stuff relating to indexed families:
\begin{spec}
{-
    id⇉ : {I : Set} {r : Indexed I} -> r ⇉ r
    id⇉ i = id
-}
    -- Merging natural transformations.
    infixr 5 _∥_
    _∥_ : {I J : Set} {r u : Indexed I} {s v : Indexed J} ->
          (r ⇉ u) -> (s ⇉ v) -> ((r ∣ s) ⇉ (u ∣ v))
    _∥_ f g (inj₁ x) = f x
    _∥_ f g (inj₂ y) = g y
\end{spec}
%endif

%endif
