{-#  OPTIONS --universe-polymorphism    #-}
{-#  OPTIONS --no-positivity-check      #-}
{-#  OPTIONS --no-termination-check     #-}
{-#  OPTIONS --type-in-type             #-}
module thesisCode where

open import Relation.Binary.PropositionalEquality public hiding ([_] ; _≗_)
open import Coinduction public
open import Data.Empty public
open import Data.Unit using (⊤ ; tt)
open import Data.Sum public hiding (map)
open import Data.Product public hiding (map ; ∃)
open import Function public
module gpintro where
data Nat : Set where
  ze : Nat
  su : Nat -> Nat
_+_ : Nat -> Nat -> Nat
m + ze      = m
m + (su n)  = su (m + n)
{-#  BUILTIN NATURAL Nat  #-}
{-#  BUILTIN ZERO ze  #-}
{-#  BUILTIN SUC su  #-}
module Type where
  data NatList : Set where
    Nil1   : NatList
    Cons1  : Nat -> NatList -> NatList
  data List  (aT : Set) : Set where
    Nil    :  (List aT)
    _∷_    :  aT -> (List aT) -> (List aT)
  length : {aT : Set} -> (List aT) -> Nat
  length Nil      = 0
  length (h ∷ t)  = 1 + length t
  map : {aT bT : Set} -> (aT -> bT) -> (List aT) -> (List bT)
  map f Nil       = Nil
  map f (h ∷ t)   = f h ∷ map f t
module Shape where
  data Tree (aT : Set) : Set where
    Leaf  : aT -> Tree aT
    Bin   : Tree aT -> Tree aT -> Tree aT
  length : {aT : Set} -> (Tree aT) -> Nat
  length (Leaf x)   = 1
  length (Bin l r)  = length l + length r

  map : {aT bT : Set} -> (aT -> bT) -> (Tree aT) -> (Tree bT)
  map f (Leaf x)   = Leaf (f x)
  map f (Bin l r)  = Bin (map f l) (map f r)
module CT where
  Functor : Set₁
  Functor = Set -> Set
  data _⊕_ (fT gT : Functor) (pT : Set) : Set where
    inj₁ : fT pT -> (fT ⊕ gT) pT
    inj₂ : gT pT -> (fT ⊕ gT) pT
  data _⊗_ (fT gT : Functor) (pT : Set) : Set where
    _,_ : fT pT -> gT pT -> (fT ⊗ gT) pT
  data Unit (pT : Set) : Set where
    unit : Unit pT
  data Bool : Set where
    true   : Bool
    false  : Bool
  BoolF : Functor
  BoolF = Unit ⊕ Unit
  data Id (pT : Set) : Set where
    id_ : pT -> Id pT
  NatF : Functor
  NatF = Unit ⊕ Id
  data μ (fT : Functor) : Set where
    ⟨_⟩ : fT (μ fT) -> μ fT
   
  out : {fT : Functor} -> μ fT -> fT (μ fT)
  out ⟨ x ⟩ = x
  data K (aT pT : Set) : Set where
    k_ : aT -> K aT pT
  ListF  : Set -> Functor
  ListF aT = Unit ⊕ (K aT ⊗ Id)
  record _≃_ (aT bT : Set) : Set where
    field
      from  : aT -> bT
      to    : bT -> aT
      iso₁  : ∀ x -> to (from x) ≡ x
      iso₂  : ∀ x -> from (to x) ≡ x
  fromNat : Nat -> μ NatF
  fromNat ze      = ⟨ inj₁ unit ⟩
  fromNat (su n)  = ⟨ inj₂ (id_ (fromNat n)) ⟩

  toNat : μ NatF -> Nat
  toNat ⟨ inj₁ unit ⟩     = ze
  toNat ⟨ inj₂ (id_ n) ⟩  = su (toNat n)
  sym1 : {aT bT : Set} -> aT ≡ bT -> bT ≡ aT
  sym1 refl = refl

  trans1 : {aT bT cT : Set} -> aT ≡ bT -> bT ≡ cT -> aT ≡ cT
  trans1 refl refl = refl

  cong1 : {aT bT : Set} {x y : aT} -> (f : aT -> bT) -> x ≡ y -> f x ≡ f y
  cong1 f refl = refl
  isoNat₁ : (n : Nat) -> toNat (fromNat n) ≡ n
  isoNat₁ ze      = refl
  isoNat₁ (su n)  = cong1 su (isoNat₁ n)

  isoNat₂ : (n : μ NatF) -> fromNat (toNat n) ≡ n
  isoNat₂ ⟨ inj₁ unit ⟩     = refl
  isoNat₂ ⟨ inj₂ (id_ n) ⟩  = cong1 (λ x -> ⟨ inj₂ (id_ x) ⟩) (isoNat₂ n)
  isoNat : Nat ≃ μ NatF
  isoNat = record  {  from  = fromNat  ;  to    = toNat
                   ;  iso₁  = isoNat₁  ;  iso₂  = isoNat₂ }
  postulate
    mapNat : {aT bT : Set} -> (aT -> bT) -> NatF aT -> NatF bT
  mapUnit : {pT pT' : Set} -> (pT -> pT') -> Unit pT -> Unit pT'
  mapUnit _ unit = unit
   
  mapSum :  {fT gT : Functor} {pT pT' : Set} ->
            (fT pT -> fT pT') -> (gT pT -> gT pT') -> (fT ⊕ gT) pT -> (fT ⊕ gT) pT'
  mapSum f g (inj₁  x) = inj₁   (f  x)
  mapSum f g (inj₂  x) = inj₂   (g  x)
   
  mapId : {pT pT' : Set} -> (pT -> pT') -> Id pT -> Id pT'
  mapId f (id_ x) = id_ (f x)
  mapNat2 : {aT bT : Set} -> (aT -> bT) -> NatF aT -> NatF bT
  mapNat2 f = mapSum (mapUnit f) (mapId f)
  cataNat : {bT : Set} -> (NatF bT -> bT) -> μ NatF -> bT
  cataNat fT = fT ∘ mapNat2 (cataNat fT) ∘ out
module Regular where
  data Code : Set where
    U'   :                 Code
    I'   :                 Code
    _⊕_  : (F G : Code) ->  Code
    _⊗_  : (F G : Code) ->  Code
  ⟦_⟧ : Code -> (Set -> Set)
  ⟦ U'     ⟧  aT = ⊤
  ⟦ I'     ⟧  aT = aT
  ⟦ F ⊕ G  ⟧  aT = ⟦ F ⟧ aT  ⊎  ⟦ G ⟧ aT
  ⟦ F ⊗ G  ⟧  aT = ⟦ F ⟧ aT  ×  ⟦ G ⟧ aT
  data μ (F : Code) : Set where
    ⟨_⟩ : ⟦ F ⟧ (μ F) -> μ F
  map  :  (F : Code) -> {aT bT : Set} -> (aT -> bT) -> ⟦ F ⟧ aT -> ⟦ F ⟧ bT

  map U'       f _          = tt
  map I'       f x          = f x
  map (F ⊕ G)  f (inj₁ x)   = inj₁ (map F f x)
  map (F ⊕ G)  f (inj₂ y)   = inj₂ (map G f y)
  map (F ⊗ G)  f (x , y)  = map F f x , map G f y
  cata' : {aT : Set} -> (F : Code) -> (⟦ F ⟧ aT -> aT) -> μ F -> aT
  cata' F f ⟨ x ⟩ = f (map F (cata' F f) x)
  NatC : Code
  NatC = U' ⊕ I'
  aNat : μ NatC
  aNat = ⟨ inj₂ ⟨ inj₂ ⟨ inj₁ tt ⟩ ⟩ ⟩
  map∀ : (C : Code) {A B : Set} {f g : A -> B} -> (∀ x -> f x ≡ g x) -> (∀ x -> map C f x ≡ map C g x)
  map∀ U' p x = refl
  map∀ I' p x = p x
  map∀ (F ⊕ G) p (inj₁ x) = cong inj₁ (map∀ F p x)
  map∀ (F ⊕ G) p (inj₂ x) = cong inj₂ (map∀ G p x)
  map∀ (F ⊗ G) p (x , y)  = cong₂ _,_ (map∀ F p x) (map∀ G p y)

  map∘ : {A B C : Set} (D : Code) {f : B -> C} {g : A -> B} {x : ⟦ D ⟧ A}
            -> map D f (map D g x) ≡ map D (f ∘ g) x
  map∘ U' = refl
  map∘ I' {_} = refl
  map∘ (F ⊕ G) {x = inj₁ _} = cong inj₁ (map∘ F)
  map∘ (F ⊕ G) {x = inj₂ _} = cong inj₂ (map∘ G)
  map∘ (F ⊗ G) {x = _ , _}  = cong₂ _,_ (map∘ F) (map∘ G)

  mapId : ∀ {A} (C : Code) {x : ⟦ C ⟧ A} -> map C id x ≡ x
  mapId U' {tt} = refl
  mapId I' = refl
  mapId (F ⊕ G) {inj₁ _} = cong inj₁ (mapId F)
  mapId (F ⊕ G) {inj₂ _} = cong inj₂ (mapId G)
  mapId (F ⊗ G) {_ , _}  = cong₂ _,_ (mapId F) (mapId G)
module PolyP where
 mutual
  data Code : Set where
    U'   :                 Code
    P'   :                 Code
    I'   :                 Code
    _⊕_  : (F G : Code) ->  Code
    _⊗_  : (F G : Code) ->  Code
    _⊚_  : (F G : Code) ->  Code
  ⟦_⟧ : Code -> (Set -> Set -> Set)
  ⟦ U'     ⟧  aT bT = ⊤
  ⟦ P'     ⟧  aT bT = aT
  ⟦ I'     ⟧  aT bT = bT
  ⟦ F ⊕ G  ⟧  aT bT = ⟦ F ⟧ aT bT ⊎  ⟦ G ⟧ aT bT
  ⟦ F ⊗ G  ⟧  aT bT = ⟦ F ⟧ aT bT ×  ⟦ G ⟧ aT bT
  ⟦ F ⊚ G  ⟧  aT bT = μ F (⟦ G ⟧ aT bT)
  data μ (F : Code) (aT : Set) : Set where
    ⟨_⟩ : ⟦ F ⟧ aT (μ F aT) -> μ F aT
  map  :  {aT bT cT dT : Set}
          (F : Code) -> (aT -> bT) -> (cT -> dT) -> ⟦ F ⟧ aT cT -> ⟦ F ⟧ bT dT
  map U'       f g tt         = tt
  map P'       f g x          = f  x
  map I'       f g x          = g  x
  map (F ⊕ G)  f g (inj₁  x)  = inj₁ (map F  f g x)
  map (F ⊕ G)  f g (inj₂  y)  = inj₂ (map G  f g y)
  map (F ⊗ G)  f g (x , y)  = map F f g x , map G f g y
  map (F ⊚ G)  f g ⟨ x ⟩      = ⟨ map F (map G f g) (map (F ⊚ G) f g) x ⟩
  pmap  :  {aT bT : Set} (F : Code) ->  (aT -> bT) -> μ F aT -> μ F bT
  pmap F f ⟨ x ⟩ = ⟨ map F f (pmap F f) x ⟩
  ListC : Code
  ListC = U' ⊕ (P' ⊗ I')
  
  RoseC : Code
  RoseC = P' ⊗ (ListC ⊚ I')
  sRose : μ RoseC ⊤
  sRose = ⟨ tt , ⟨ inj₁ tt ⟩ ⟩ 
  
  lRose : μ RoseC ⊤
  lRose = ⟨ tt , ⟨ inj₂ (sRose , ⟨ inj₂ (sRose , ⟨ inj₁ tt ⟩)⟩)⟩ ⟩
module Multirec where
  Indexed : Set -> Set₁
  Indexed aT = aT -> Set
  data Code (I : Set) : Set where
    U'   :                    Code I
    I'   : I               ->  Code I
    !    : I               ->  Code I
    _⊕_  : (F G : Code I)  ->  Code I
    _⊗_  : (F G : Code I)  ->  Code I
  ⟦_⟧ : {I : Set} -> Code I -> Indexed I -> Indexed I
  ⟦ U'     ⟧  r i = ⊤
  ⟦ I' j   ⟧  r i = r j
  ⟦ ! j    ⟧  r i = i ≡ j
  ⟦ F ⊕ G  ⟧  r i = ⟦ F ⟧ r i  ⊎  ⟦ G ⟧ r i
  ⟦ F ⊗ G  ⟧  r i = ⟦ F ⟧ r i  ×  ⟦ G ⟧ r i
  _⇉_ : {I : Set} -> Indexed I -> Indexed I -> Set
  r ⇉ s = ∀ i -> r i -> s i
  map  : {I : Set} {r s : Indexed I} (F : Code I) -> r ⇉ s -> ⟦ F ⟧ r ⇉ ⟦ F ⟧ s
  map U'       f i tt          = tt
  map (I' j)   f i x           = f j x
  map (! j)    f i x           = x
  map (F ⊕ G)  f i (inj₁ x)    = inj₁  (map F  f i x)
  map (F ⊕ G)  f i (inj₂ y)    = inj₂  (map G  f i y)
  map (F ⊗ G)  f i (x , y)   = map F f i x , map G f i y
  data μ {I : Set} (F : Code I) (i : I) : Set where
    ⟨_⟩ : ⟦ F ⟧ (μ F) i -> μ F i
  mutual
    data Zig :  Set where
      zig  : Zag -> Zig
      end  : Zig
    
    data Zag :  Set where
      zag : Zig -> Zag
  ZigC  : Code (⊤ ⊎ ⊤)
  ZigC    = I' (inj₂ tt) ⊕ U'

  ZagC  : Code (⊤ ⊎ ⊤)
  ZagC    = I' (inj₁ tt)

  ZigZagC : Code (⊤ ⊎ ⊤)
  ZigZagC = (! (inj₁ tt) ⊗ ZigC) ⊕ (! (inj₂ tt) ⊗ ZagC)
  
  zigZagEnd : μ ZigZagC (inj₁ tt)
  zigZagEnd = ⟨ inj₁ (refl , inj₁ ⟨ inj₂ (refl , ⟨ inj₁ (refl , inj₂ tt) ⟩) ⟩) ⟩
module IndexedOrig where

  open import appendix
  module univ1 where
    data _▸_ (I : Set) (O : Set) : Set₁ where
      Z'   : I ▸ O
      U'   : I ▸ O
      _⊕_  :               I  ▸ O  -> I ▸  O  -> I ▸ O
      _⊗_  :               I  ▸ O  -> I ▸  O  -> I ▸ O
      _⊚_  : {M : Set} ->  M  ▸ O  -> I ▸  M  -> I ▸ O
    ⟦_⟧ : {I O : Set} -> I ▸ O -> I ▷ O
    ⟦ Z'     ⟧  r o =  ⊥
    ⟦ U'     ⟧  r o =  ⊤
    ⟦ F ⊕ G  ⟧  r o =  ⟦ F ⟧ r o  ⊎  ⟦ G ⟧ r o
    ⟦ F ⊗ G  ⟧  r o =  ⟦ F ⟧ r o  ×  ⟦ G ⟧ r o
    ⟦ F ⊚ G  ⟧  r o =  (⟦ F ⟧ ∘ ⟦ G ⟧) r o
  module univ2 where
    mutual
      data _▸_ : Set -> Set -> Set₁ where
        Z'   : ∀ {I O}                 -> I ▸ O
        U'   : ∀ {I O}                 -> I ▸ O
        I'   : ∀ {I O}    -> (i : I)    -> I ▸ O
        _⊕_  : ∀ {I O}    -> (F : I ▸ O)  -> (G : I ▸ O)  -> I ▸ O
        _⊗_  : ∀ {I O}    -> (F : I ▸ O)  -> (G : I ▸ O)  -> I ▸ O
        _⊚_  : ∀ {I M O}  -> (F : M ▸ O)  -> (G : I ▸ M)  -> I ▸ O
        Iso  : ∀ {I O} -> (C : I ▸ O) -> (D : I ▷ O)
                       -> (e : (r : Indexed I) -> (o : O) -> D r o ≃ ⟦ C ⟧ r o)
                       -> I ▸ O

      ⟦_⟧ : ∀ {I O} -> I ▸ O -> I ▷ O
      ⟦ Z'    ⟧  r o =  ⊥
      ⟦ U'    ⟧  r o =  ⊤
      ⟦ I' i  ⟧  r o =  r i
      ⟦ F ⊕ G ⟧  r o =  ⟦ F ⟧ r o ⊎ ⟦ G ⟧ r o
      ⟦ F ⊗ G ⟧  r o =  ⟦ F ⟧ r o × ⟦ G ⟧ r o
      ⟦ F ⊚ G ⟧  r o =  ⟦ F ⟧ (⟦ G ⟧ r) o
      ⟦ Iso C D e ⟧ r o = D r o
    data Bool : Set where
      true   : Bool
      false  : Bool
    ‵BoolC′ : ⊥ ▸ ⊤
    ‵BoolC′ = U' ⊕ U'
    fromBool : Bool -> ⟦ ‵BoolC′ ⟧ (\ ()) tt
    fromBool  true   = inj₁ tt
    fromBool  false  = inj₂ tt

    toBool : ⟦ ‵BoolC′ ⟧ (\ ()) tt -> Bool
    toBool  (inj₁  tt) =  true
    toBool  (inj₂  tt) =  false
    isoBool₁  : ∀ b -> toBool (fromBool b) ≡ b
    isoBool₁ true   = refl
    isoBool₁ false  = refl

    isoBool₂  : ∀ b -> fromBool (toBool b) ≡ b
    isoBool₂ (inj₁  tt)  = refl
    isoBool₂ (inj₂  tt)  = refl
    isoBool : (r : Indexed ⊥) (o : ⊤) -> Bool ≃ ⟦ ‵BoolC′ ⟧ r o
    isoBool r o = record  {  from  = fromBool
                          ;  to    = toBool
                          ;  iso₁  = isoBool₁
                          ;  iso₂  = isoBool₂ }
    ‵Bool′ : ⊥ ▸ ⊤
    ‵Bool′ = Iso ‵BoolC′ (λ _ _ -> Bool) isoBool
    ‵ListC′ : (⊤ ⊎ ⊤) ▸ ⊤
    ‵ListC′ = U' ⊕ (I' (inj₁ tt) ⊗ I' (inj₂ tt))
  module univ3 where

    infixr 3 _⊕_
    infixr 4 _⊗_
    infixr 5 _⊚_

    mutual
      data _▸_ : Set -> Set -> Set₁ where
        Z'   : ∀ {I O}                 -> I ▸ O
        U'   : ∀ {I O}                 -> I ▸ O
        I'   : ∀ {I O}    -> (i : I)    -> I ▸ O
        _⊕_  : ∀ {I O}    -> (F : I ▸ O)  -> (G : I ▸ O)  -> I ▸ O
        _⊗_  : ∀ {I O}    -> (F : I ▸ O)  -> (G : I ▸ O)  -> I ▸ O
        _⊚_  : ∀ {I M O}  -> (F : M ▸ O)  -> (G : I ▸ M)  -> I ▸ O
        Fix' : ∀ {I O}    -> (I ⊎ O) ▸ O -> I ▸ O
        Iso  : ∀ {I O}    -> (C : I ▸ O) -> (D : I ▷ O)
                          -> (e : (r : Indexed I) -> (o : O) -> D r o ≃ ⟦ C ⟧ r o)
                          -> I ▸ O

      data μ {I O : Set} (F : (I ⊎ O) ▸ O) (r : Indexed I) (o : O) : Set where
        ⟨_⟩ : ⟦ F ⟧ (r ∣ μ F r) o -> μ F r o

      ⟦_⟧ : ∀ {I O} -> I ▸ O -> I ▷ O
      ⟦ Z'    ⟧  r o =  ⊥
      ⟦ U'    ⟧  r o =  ⊤
      ⟦ I' i  ⟧  r o =  r i
      ⟦ F ⊕ G ⟧  r o =  ⟦ F ⟧ r o ⊎ ⟦ G ⟧ r o
      ⟦ F ⊗ G ⟧  r o =  ⟦ F ⟧ r o × ⟦ G ⟧ r o
      ⟦ F ⊚ G ⟧  r o =  ⟦ F ⟧ (⟦ G ⟧ r) o
      ⟦ Fix' F ⟧  r o =  μ F r o
      ⟦ Iso C D e ⟧ r o = D r o

    ‵ℕF′ : (⊥ ⊎ ⊤) ▸ ⊤
    ‵ℕF′ = U' ⊕ I' (inj₂ tt)

    infixr 6 _∷_
    data [_] (A : Set) : Set where
      _∷_ : (h : A) -> (t : [ A ]) -> [ A ]
      []  : [ A ]

    List : Set -> Set
    List = [_]

    ‵ListC′ : (⊤ ⊎ ⊤) ▸ ⊤
    ‵ListC′ = U' ⊕ (I' (inj₁ tt) ⊗ I' (inj₂ tt))
    ‵ListF′ : ⊤ ▸ ⊤
    ‵ListF′ = Fix' ‵ListC′
    fromList : ∀ {r o} -> (List (r o)) -> ⟦ ‵ListF′ ⟧ r o
    fromList []       = ⟨ inj₁  tt ⟩
    fromList (x ∷ xs)  = ⟨ inj₂  (x , fromList xs) ⟩

    toList : ∀ {r o} -> ⟦ ‵ListF′ ⟧ r o -> (List (r o))
    toList ⟨ inj₁ tt ⟩          = []
    toList ⟨ inj₂ (x , xs) ⟩  = x ∷ toList xs
    isoList₁ :  ∀ {r o} -> (l : (List (r o))) -> toList {r} (fromList l) ≡ l
    isoList₁      []      = refl
    isoList₁ {r}  (h ∷ t)  = cong (λ x -> h ∷ x) (isoList₁ {r} t)

    isoList₂ :  ∀ {r o} -> (l : ⟦ ‵ListF′ ⟧ r o) -> fromList (toList l) ≡ l
    isoList₂   ⟨ inj₁ tt ⟩           = refl
    isoList₂   ⟨ inj₂ (h , t) ⟩    = cong (λ x -> ⟨ inj₂ (h , x) ⟩) (isoList₂ t)

    isoList : ∀ {r o} -> (List (r o)) ≃ ⟦ ‵ListF′ ⟧ r o
    isoList {r} = record  {  from  = fromList
                          ;  to    = toList
                          ;  iso₁  = isoList₁ {r}
                          ;  iso₂  = isoList₂ }
    ‵List′ : ⊤ ▸ ⊤
    ‵List′ = Iso (Fix' ‵ListC′) (λ f t -> (List (f t))) (λ r o -> isoList)
    map :  {I O : Set} {r s : Indexed I} ->
           (C : I ▸ O) -> r ⇉ s -> ⟦ C ⟧ r ⇉ ⟦ C ⟧ s
    map Z'         f o  ()
    map U'         f o  tt                  = tt
    map (I' i)     f o  x                   = f i x
    map (F ⊕ G)    f o  (inj₁  x)           = inj₁  (map F  f o x)
    map (F ⊕ G)    f o  (inj₂  x)           = inj₂  (map G  f o x)
    map (F ⊗ G)    f o  (x , y)           = map F f o x , map G f o y
    map (F ⊚ G)    f o  x                   = map F (map G f) o x
    map (Fix' F)   f o  ⟨ x ⟩               = ⟨ map F (f ∥ map (Fix' F) f) o x ⟩
    map {r = r} {s = s} (Iso C D e) f o x with (e r o , e s o)
    ... | ep₁ , ep₂ = _≃_.to ep₂ (map C f o (_≃_.from ep₁ x))
    mapList : {AT BT : Set} -> (AT -> BT) -> (List AT) -> (List BT)
    mapList f = map ‵List′ (const f) tt
    mapListExample : mapList su (1 ∷ 2 ∷ []) ≡ (2 ∷ 3 ∷ [])
    mapListExample = refl
    id⇉ : {I : Set} {r : Indexed I} -> r ⇉ r
    id⇉ i = id

    cata' :  {I O : Set} {r : Indexed I} {s : Indexed O} -> 
             (C : (I ⊎ O) ▸ O) -> ⟦ C ⟧ (r ∣ s) ⇉ s -> ⟦ Fix' C ⟧ r ⇉ s
    cata' C φ i ⟨ x ⟩ = φ i (map C (id⇉ ∥ cata' C φ) i x)
    _▿_ : {AT BT CT : Set} -> (AT -> CT) -> (BT -> CT) -> (AT ⊎ BT) -> CT
    (r ▿ s) (inj₁ i)  =  r  i
    (r ▿ s) (inj₂ j)  =  s  j

    foldr : {AT BT : Set} -> (AT -> BT -> BT) -> BT -> (List AT) -> BT
    foldr {AT} c n l =  cata' {r = const AT} ‵ListC′ φ tt (fromList l)
                        where φ = const (const n ▿ uncurry c)
    length : {AT : Set} -> (List AT) -> Nat
    length = foldr (const su) ze

    lengthExample : length (1 ∷ 0 ∷ []) ≡ 2
    lengthExample = refl
    ana   :  {I O : Set} {r : Indexed I} {s : Indexed O} ->
             (C : (I ⊎ O) ▸ O) -> s ⇉ ⟦ C ⟧ (r ∣ s) -> s ⇉ ⟦ Fix' C ⟧ r

    ana   C     ψ  i  x      = ⟨ map C (id⇉ ∥ ana C ψ) i (ψ i x) ⟩
     
    hylo  :  {I O : Set} {r : Indexed I} {s t : Indexed O} ->
             (C : (I ⊎ O) ▸ O) -> ⟦ C ⟧ (r ∣ t) ⇉ t -> s ⇉ ⟦ C ⟧ (r ∣ s) -> s ⇉ t

    hylo  C  φ  ψ  i  x      = φ i (map C (id⇉ ∥ hylo C φ ψ) i (ψ i x))
    data Rose (AT : Set) : Set where
      fork : AT -> (List (Rose AT)) -> Rose AT
    ‵RoseC′ : (⊤ ⊎ ⊤) ▸ ⊤
    ‵RoseC′ = I' (inj₁ tt) ⊗ (‵ListF′ ⊚ I' (inj₂ tt))

    ‵RoseF′ : ⊤ ▸ ⊤
    ‵RoseF′ = Fix' ‵RoseC′
    fromRose :  {r : Indexed ⊤} {o : ⊤} ->  Rose (r o) -> ⟦ ‵RoseF′ ⟧ r o
    fromRose (fork x xs) = ⟨ x , map ‵ListF′ (λ i -> fromRose) tt (fromList xs) ⟩

    toRose :  {r : Indexed ⊤} {o : ⊤} ->  ⟦ ‵RoseF′ ⟧ r o -> Rose (r o)
    toRose ⟨ x , xs ⟩ = fork x (toList (map ‵ListF′ (λ i -> toRose) tt xs))
    r1 : Rose Nat
    r1 = fork 3 (fork 2 [] ∷ fork 1 [] ∷ [])

    test : ⟦ ‵RoseF′ ⟧ (const Nat) tt
    test = fromRose r1
  module univ4 where

    infixr 3 _⊕_
    infixr 4 _⊗_
    infixr 5 _⊚_

    mutual
      data _▸_ : Set -> Set -> Set₁ where
        Z'   : ∀ {I O}                 -> I ▸ O
        U'   : ∀ {I O}                 -> I ▸ O
        I'   : ∀ {I O}    -> (i : I)    -> I ▸ O
        _⊕_  : ∀ {I O}    -> (F : I ▸ O)  -> (G : I ▸ O)  -> I ▸ O
        _⊗_  : ∀ {I O}    -> (F : I ▸ O)  -> (G : I ▸ O)  -> I ▸ O
        _⊚_  : ∀ {I M O}  -> (F : M ▸ O)  -> (G : I ▸ M)  -> I ▸ O
        _↗_↘_ : ∀ {I I′ O O′} -> (F : I′ ▸ O′) -> (I′ -> I) -> (O -> O′) -> I ▸ O
        Fix' : ∀ {I O}    -> (I ⊎ O) ▸ O -> I ▸ O
        !    : ∀ {I O}    -> (o′ : O)    -> I ▸ O
--         Σ    : ∀ {I O}    -> {A : Set}       -> (A -> I ▸ O)  -> I ▸ O
        Sig  : ∀ {I O} {C : ⊥ ▸ ⊤} -> (⟦ C ⟧ (λ _ -> ⊤) tt -> I ▸ O) -> I ▸ O
        Iso  : ∀ {I O}    -> (C : I ▸ O) -> (D : I ▷ O)
                          -> (e : (r : Indexed I) -> (o : O) -> D r o ≃ ⟦ C ⟧ r o)
                          -> I ▸ O

      data μ {I O : Set} (F : (I ⊎ O) ▸ O) (r : Indexed I) (o : O) : Set where
        ⟨_⟩ : ⟦ F ⟧ (r ∣ μ F r) o -> μ F r o

      ⟦_⟧ : ∀ {I O} -> I ▸ O -> I ▷ O
      ⟦ Z'    ⟧  r o =  ⊥
      ⟦ U'    ⟧  r o =  ⊤
      ⟦ I' i  ⟧  r o =  r i
      ⟦ F ↗ f ↘ g ⟧  r o = ⟦ F ⟧ (r ∘ f) (g o)
      ⟦ F ⊕ G ⟧  r o =  ⟦ F ⟧ r o ⊎ ⟦ G ⟧ r o
      ⟦ F ⊗ G ⟧  r o =  ⟦ F ⟧ r o × ⟦ G ⟧ r o
      ⟦ F ⊚ G ⟧  r o =  ⟦ F ⟧ (⟦ G ⟧ r) o
      ⟦ Fix' F ⟧ r o =  μ F r o
      ⟦ ! o′  ⟧  r o =  o ≡ o′
      ⟦ Sig f ⟧  r o =  ∃ (λ i -> ⟦ f i ⟧ r o)
      ⟦ Iso C D e ⟧ r o = D r o

    map : {I O : Set} {r s : Indexed I} (C : I ▸ O) ->
          (r ⇉ s) -> (⟦ C ⟧ r ⇉ ⟦ C ⟧ s)
    map Z'         f o ()
    map U'         f o x                   = x
    map (I' i)     f o x                   = f i x
    map (F ⊕ G)    f o (inj₁ x)             = inj₁ (map F f o x)
    map (F ⊕ G)    f o (inj₂ y)             = inj₂ (map G f o y)
    map (F ⊗ G)    f o (x , y)             = map F f o x , map G f o y
    map (F ↗ g ↘ h) f o x                  = map F (f ∘ g) (h o) x
    map (F ⊚ G)    f o x                   = map F (map G f) o x
    map (! o')     f o x                   = x
    map (Sig g)    f o (some {i} x)        = some (map (g i) f o x)
    map (Fix' F)   f o ⟨ x ⟩               = ⟨ map F (f ∥ map (Fix' F) f) o x ⟩
    map {r = r} {s = s} (Iso C D e) f o x with (e r o , e s o)
    ... | (ep₁ , ep₂) = _≃_.to ep₂ (map C f o (_≃_.from ep₁ x))

    ‵NatF′ : (⊥ ⊎ ⊤) ▸ ⊤
    ‵NatF′ = U' ⊕ I' (inj₂ tt)

    ‵ListC′ : (⊤ ⊎ ⊤) ▸ ⊤
    ‵ListC′ = U' ⊕ (I' (inj₁ tt) ⊗ I' (inj₂ tt))

    ‵ListF′ : ⊤ ▸ ⊤
    ‵ListF′ = Fix' ‵ListC′

    fromNat : ∀ {r o} -> Nat -> ⟦ Fix' ‵NatF′ ⟧ r o
    fromNat ze      = ⟨ inj₁ tt ⟩
    fromNat (su n)   = ⟨ inj₂ (fromNat n) ⟩

    toNat : ∀ {r o} -> ⟦ Fix' ‵NatF′ ⟧ r o -> Nat
    toNat ⟨ inj₁ tt ⟩  = ze
    toNat ⟨ inj₂ n ⟩   = su (toNat n)

    toFromNat≃ : ∀ {r o} (n : Nat) -> toNat {r} {o} (fromNat n) ≡ n
    toFromNat≃ ze  = refl
    toFromNat≃ (su n) = cong su (toFromNat≃ n)

    fromToNat≃ : ∀ {r o} (x : ⟦ Fix' ‵NatF′ ⟧ r o) -> (fromNat (toNat x)) ≡ x
    fromToNat≃ ⟨ inj₁ tt ⟩ = refl
    fromToNat≃ ⟨ inj₂ _  ⟩ = cong (⟨_⟩ ∘ inj₂) (fromToNat≃ _)

    ‵ℕ′ : ⊥ ▸ ⊤
    ‵ℕ′ = Iso (Fix' ‵NatF′) (λ _ _ -> Nat)
              (λ r o -> record { from = fromNat {r} {o}
                               ; to   = toNat {r} {o}
                               ; iso₁  = toFromNat≃ {r} {o} 
                               ; iso₂  = fromToNat≃ {r} {o} } )
    mutual
      data Expr  (AT : Set)  : Set where
        econst  : Nat                   -> Expr AT
        add     : Expr AT  -> Expr AT   -> Expr AT
        evar    : AT                    -> Expr AT
        elet    : Decl AT  -> Expr AT   -> Expr AT

      data Decl  (AT : Set)  : Set where
        assign  : AT       -> Expr AT  -> Decl AT
        seq     : Decl AT  -> Decl AT  -> Decl AT
    AST : Set
    AST = ⊤ ⊎ ⊤

    expr  : AST
    expr  = inj₁ tt

    decl  : AST
    decl  = inj₂ tt
    ‵ExprC′ : (⊤ ⊎ AST) ▸ AST
    ‵ExprC′  =     (‵ℕ′ ↗ (λ ()) ↘ const tt)
                ⊕  I' (inj₂  expr)  ⊗ I' (inj₂ expr)
                ⊕  I' (inj₁  tt)
                ⊕  I' (inj₂  decl)  ⊗ I' (inj₂ expr)

    ‵DeclC′ : (⊤ ⊎ AST) ▸ AST
    ‵DeclC′  =     I' (inj₁  tt)    ⊗ I' (inj₂ expr)
                ⊕  I' (inj₂  decl)  ⊗ I' (inj₂ decl)
    ‵ASTC′  : (⊤ ⊎ AST) ▸ AST
    ‵ASTC′  =  !  expr  ⊗  ‵ExprC′
            ⊕  !  decl  ⊗  ‵DeclC′

    ‵ASTF′ : ⊤ ▸ AST
    ‵ASTF′ = Fix' ‵ASTC′
    mutual
      toExpr : {r : ⊤ -> Set} -> ⟦ ‵ASTF′ ⟧ r expr -> Expr (r tt)
      toExpr ⟨ inj₁  (refl  , inj₁ x) ⟩                         = econst x
      toExpr ⟨ inj₁  (refl  , inj₂ (inj₁ (x , y))) ⟩          = add (toExpr x) (toExpr y)
      toExpr ⟨ inj₁  (refl  , inj₂ (inj₂ (inj₁ x))) ⟩           = evar x
      toExpr ⟨ inj₁  (refl  , inj₂ (inj₂ (inj₂ (d , e)))) ⟩   = elet (toDecl d) (toExpr e)
      toExpr ⟨ inj₂  (()    , _) ⟩

      toDecl : {r : ⊤ -> Set} -> ⟦ ‵ASTF′ ⟧ r decl -> Decl (r tt)
      toDecl ⟨  inj₁ (()    , _) ⟩
      toDecl ⟨  inj₂ (refl  , inj₁ (x , e)) ⟩    = assign x (toExpr e)
      toDecl ⟨  inj₂ (refl  , inj₂ (d₁ , d₂)) ⟩  = seq  (toDecl d₁) (toDecl d₂)
    mutual
      fromExpr :  {r : ⊤ -> Set} -> Expr (r tt) -> ⟦ ‵ASTF′ ⟧ r expr
      fromExpr (econst x)  = ⟨ inj₁ (refl , inj₁ x) ⟩
      fromExpr (add x y)   = ⟨ inj₁ (refl , inj₂ (inj₁ (fromExpr x , fromExpr y))) ⟩
      fromExpr (evar x)    = ⟨ inj₁ (refl , inj₂ (inj₂ (inj₁ x))) ⟩
      fromExpr (elet d e)  = ⟨ inj₁ (refl , inj₂ (inj₂ (inj₂ (fromDecl d , fromExpr e))))⟩

      fromDecl :  {r : ⊤ -> Set} -> Decl (r tt) -> ⟦ ‵ASTF′ ⟧ r decl
      fromDecl (assign x e)  = ⟨ inj₂ (refl , inj₁ (x , fromExpr e)) ⟩
      fromDecl (seq d₁ d₂)   = ⟨ inj₂ (refl , inj₂ (fromDecl d₁ , fromDecl d₂)) ⟩
    infixr 5 _∷_

    data Vec (AT : Set) : Nat -> Set where
      []  :                                 Vec AT  ze
      _∷_  : {n : Nat} -> AT -> Vec AT n ->  Vec AT  (su n)
    ‵VecC′ : Nat -> (⊤ ⊎ Nat) ▸ Nat
    ‵VecC′  ze       = !  ze
    ‵VecC′  (su n)   = !  (su n) ⊗ I' (inj₁ tt) ⊗ I' (inj₂ n)
    ‵VecF′ : ⊤ ▸ Nat
    ‵VecF′ = Fix' (Sig {C = ‵ℕ′} ‵VecC′)
    fromVec  :  ∀ {n r} -> Vec (r tt) n ->  ⟦ ‵VecF′ ⟧ r n
    fromVec {n = ze}    []      = ⟨ some  {x = ze}    refl ⟩
    fromVec {n = su m}  (h ∷ t)  = ⟨ some  {x = su m}  (refl , (h , fromVec t)) ⟩

    toVec  : ∀ {n r} -> ⟦ ‵VecF′ ⟧ r n -> Vec (r tt) n
    toVec ⟨ some  {ze}    refl ⟩                  =  []
    toVec ⟨ some  {su n}  (refl , (h , t)) ⟩  =  h ∷ toVec t
    data Perfect (AT : Set) : {n : Nat} -> Set where
      split  :  {n : Nat} -> Perfect AT {n} × Perfect AT {n} -> Perfect AT {su n}
      leaf   :  AT -> Perfect AT {ze}
    data Pair (AT : Set) : Set where
      pair : AT -> AT -> Pair AT

    data Perfect' : Set -> Set₁ where
      split  : {  AT  : Set  }  -> Perfect' (Pair AT)  -> Perfect' AT
      leaf   : {  AT  : Set  }  -> AT                  -> Perfect' AT
    ‵PerfectC′ : Nat -> (⊤ ⊎ Nat) ▸ Nat
    ‵PerfectC′ (ze)     = !  ze      ⊗ I' (inj₁ tt)
    ‵PerfectC′ (su n)   = !  (su n)  ⊗ I' (inj₂ n) ⊗ I' (inj₂ n)

    ‵PerfectF′ : ⊤ ▸ Nat
    ‵PerfectF′ = Fix' (Sig {C = ‵ℕ′} ‵PerfectC′)
  module indexedSummary where

    open import appendix

    infixr 3 _⊕_
    infixr 4 _⊗_
    infixr 5 _⊚_

    mutual
      data _▸_ (I : Set) (O : Set) : Set₁ where
        Z'     :  I ▸ O
        U'     :  I ▸ O
                  
        I'     :  I  -> I ▸ O
        !      :  O  -> I ▸ O
                  
        _⊕_    :             I  ▸ O  -> I ▸ O  ->  I ▸ O
        _⊗_    :             I  ▸ O  -> I ▸ O  ->  I ▸ O
        _⊚_    :  ∀ {M} ->   M  ▸ O  -> I ▸ M  ->  I ▸ O

        Fix'   :  (I ⊎ O) ▸ O -> I ▸ O

        _↗_↘_  :  ∀ {I′ O′} -> I′ ▸ O′ -> (I′ -> I) -> (O -> O′)      ->  I ▸ O

        Sig    :  {C : ⊥ ▸ ⊤} -> (⟦ C ⟧ (const ⊤) tt -> I ▸ O)        -> I ▸ O

        Iso    :  (C : I ▸ O) -> (D : I ▷ O) -> 
                  ((r : Indexed I) -> (o : O) -> D r o ≃ ⟦ C ⟧ r o)   -> I ▸ O
       
      data μ  {I O : Set} (F : (I ⊎ O) ▸ O) (r : Indexed I) (o : O) : Set where
        ⟨_⟩ : ⟦ F ⟧ (r ∣ μ F r) o -> μ F r o
       
      ⟦_⟧ : ∀ {I O} -> I ▸ O -> I ▷ O
      ⟦ Z'          ⟧      r o =  ⊥
      ⟦ U'          ⟧      r o =  ⊤
      ⟦ I' i        ⟧      r o =  r i
      ⟦ F ↗ f ↘ g   ⟧      r o =  ⟦ F ⟧ (r ∘ f) (g o)
      ⟦ F ⊕ G       ⟧      r o =  ⟦ F ⟧ r o ⊎   ⟦ G ⟧ r o
      ⟦ F ⊗ G       ⟧      r o =  ⟦ F ⟧ r o ×   ⟦ G ⟧ r o
      ⟦ F ⊚ G       ⟧      r o =  ⟦ F ⟧ (⟦ G ⟧ r) o
      ⟦ Fix' F      ⟧      r o =  μ F r o
      ⟦ ! o′        ⟧      r o =  o ≡ o′
      ⟦ Sig f       ⟧      r o =  ∃ (λ i -> ⟦ f i ⟧ r o)
      ⟦ Iso C D e   ⟧      r o =  D r o
       
      map :  {I O : Set} {r s : Indexed I}
             (C : I ▸ O) -> (r ⇉ s) -> (⟦ C ⟧ r ⇉ ⟦ C ⟧ s)
      map Z'           f o ()
      map U'           f o x                   =  x
      map (I' i)       f o x                   =  f i x
      map (F ⊕ G)      f o (inj₁   x)          =  inj₁  (map F  f o x)
      map (F ⊕ G)      f o (inj₂   x)          =  inj₂  (map G  f o x)
      map (F ⊗ G)      f o (x , y)           =  map F f o x , map G f o y
      map (F ↗ g ↘ h)  f o x                   =  map F (f ∘ g) (h o) x
      map (F ⊚ G)      f o x                   =  map F (map G f) o x
      map (! o')       f o x                   =  x
      map (Sig g)      f o (some {i} x)        =  some (map (g i) f o x)
                                                  
      map (Fix' F)     f o ⟨ x ⟩               =  ⟨ map F (f ∥ map (Fix' F) f) o x ⟩
                                                  
      map {r = r} {s = s} (Iso C D e) f o x with  (e r o , e s o)
      ... | (ep₁ , ep₂)                      =  to ep₂ (map C f o (from ep₁ x))
                                                  where open _≃_
  module proofs where

    open import zipperAux
    open import appendix
    open ≡-Reasoning
    _⇉∘_ : {I : Set} {r s t : Indexed I} -> s ⇉ t -> r ⇉ s -> r ⇉ t
    (f ⇉∘ g) i x = f i (g i x)
     
    infix 4 _≗_
    _≗_ : ∀ {I : Set}{r s : Indexed I} (f g : r ⇉ s) -> Set
    f ≗ g = ∀ i x -> f i x ≡ g i x
    ∥-cong : {I J : Set}{r u : Indexed I}{s v : Indexed J}{f₁ f₂ : r ⇉ u}{g₁ g₂ : s ⇉ v} ->
             f₁ ≗ f₂ -> g₁ ≗ g₂ -> f₁ ∥ g₁ ≗ f₂ ∥ g₂
    ∥-cong if ig (inj₁ i) x = if i x
    ∥-cong if ig (inj₂ j) x = ig j x

    ∥-id : {I J : Set} {r : Indexed I} {s : Indexed J}{f : r ⇉ r}{g : s ⇉ s} ->
           f ≗ id⇉ -> g ≗ id⇉ -> (f ∥ g) ≗ id⇉
    ∥-id if ig (inj₁ i) x = if i x
    ∥-id if ig (inj₂ j) x = ig j x
{- 
    ∥-∘ : {I J : Set}{r₁ s₁ t₁ : Indexed I}{r₂ s₂ t₂ : Indexed J} -> 
          {f₁ : s₁ ⇉ t₁}{g₁ : r₁ ⇉ s₁}{f₂ : s₂ ⇉ t₂}{g₂ : r₂ ⇉ s₂} ->
          (f₁ ⇉∘ g₁ ∥ f₂ ⇉∘ g₂) ≗ ((f₁ ∥ f₂) ⇉∘ (g₁ ∥ g₂))
    ∥-∘ (inj₁ i) x = refl
    ∥-∘ (inj₂ j) x = refl
 -}
    map-cong  :  {I O : Set} {r s : Indexed I} {f g : r ⇉ s}
                 (C : I ▸ O) -> f ≗ g -> map C f ≗ map C g

    map-cong Z'           ip i ()
    map-cong U'           ip i x          = refl
    map-cong (I' i′)      ip i x          = ip i′ x
    map-cong (F ↗ f ↘ g)  ip i x          = map-cong F (ip ∘ f) (g i) x
    map-cong (F ⊕ G)      ip i (inj₁  x)   = cong inj₁  (map-cong F  ip i x)
    map-cong (F ⊕ G)      ip i (inj₂  x)   = cong inj₂  (map-cong G  ip i x)
    map-cong (F ⊗ G)      ip i (x , y)   = cong₂ _,_  (map-cong F  ip i x) (map-cong G  ip i y)
    map-cong (F ⊚ G)      ip i x           = map-cong F (map-cong G ip) i x
    map-cong (! o′)       ip i x           = refl
    map-cong (Sig g)   ip i (some x)   = cong some  (map-cong (g _) ip i x)

    map-cong (Fix' F)  ip i ⟨ x ⟩      = cong ⟨_⟩   (map-cong F (∥-cong ip (map-cong (Fix' F) ip)) i x)

    map-cong {r = r} {s = s} (Iso C D e) ip i x =
      cong (to (e s i)) (map-cong C ip i (from (e r i) x)) where open _≃_
    map-id :  {I O : Set}{r : Indexed I}(C : I ▸ O) ->
              map {r = r} C id⇉ ≗ id⇉

    map-id Z'           i ()
    map-id U'           i x               = refl
    map-id (I' i′)      i x               = refl
    map-id (F ↗ f ↘ g)  i x               = map-id F (g i) x
    map-id (F ⊕ G)      i (inj₁ x)        = cong inj₁ (map-id F i x)
    map-id (F ⊕ G)      i (inj₂ x)        = cong inj₂ (map-id G i x)
    map-id (F ⊗ G)      i (x , y)       = cong₂ _,_ (map-id F i x) (map-id G i y)
    map-id (F ⊚ G)      i x               = trans (map-cong F (map-id G) i x) (map-id F i x)
    map-id (! o′)       i x               = refl
    map-id (Sig g)      i (some x)        = cong some (map-id (g _) i x)
    map-id (Fix' F)     i ⟨ x ⟩           = cong ⟨_⟩ (trans  (map-cong F (∥-id  (λ _ _ -> refl)
                                                                                (map-id (Fix' F))) i x)
                                                             (map-id F i x))
    map-id {r = r} (Iso C D e) i x =

      begin

        to (e r i) (map C id⇉ i (from (e r i) x))

      ≡⟨ cong (to (e r i)) (map-id C i (from (e r i) x)) ⟩

        to (e r i) (from (e r i) x)

      ≡⟨ iso₁ (e r i) x ⟩

        x ∎

      where open _≃_
  module deq where
    open import appendix
    open import zipperAux renaming (Nat to Nat2)

    List : Set -> Set
    List = [_]
    SemiDec : ∀ {I} -> Indexed I -> Set
    SemiDec r = ∀ i -> (x y : r i) -> Maybe (x ≡ y)
    infixr 5 _∦_

    _∦_ :  {I J : Set} {r : Indexed I} {s : Indexed J} ->
           SemiDec r -> SemiDec s -> SemiDec (r ∣ s)
    (f ∦ g) (inj₁  i)  = f  i
    (f ∦ g) (inj₂  i)  = g  i
    deqt :  {I O : Set} {r : Indexed I} (C : I ▸ O) ->
            SemiDec r -> SemiDec (⟦ C ⟧ r)
    deqt  Z'      f o  ()  y
    deqt  U'      f o  tt  tt  = just refl
    deqt  (I' i)  f o  x   y   = f i x y
    deqt (F ↗ f ↘ g) h o x y = deqt F (h ∘ f) (g o) x y
    deqt (F ⊕ G) f o (inj₁  x) (inj₂  y) =  nothing
    deqt (F ⊕ G) f o (inj₂  x) (inj₁  y) =  nothing

    deqt (F ⊕ G) f o (inj₁  x) (inj₁  y) =  deqt F  f o x y >>= just ∘ cong  inj₁
    deqt (F ⊕ G) f o (inj₂  x) (inj₂  y) =  deqt G  f o x y >>= just ∘ cong  inj₂
    deqt (F ⊗ G) f o (x₁ , x₂) (y₁ , y₂)  =         deqt F  f o x₁  y₁  >>= 
                                              \  l  ->  deqt G  f o x₂  y₂  >>= 
                                              \  r  ->  just (cong₂ _,_ l r)
    deqt (F ⊚ G)  f  o   x     y     = deqt F (deqt G f) o x y 
    deqt (! o)    f  .o  refl  refl  = just refl
    deqt (Sig {C = C} g) f o (some {i₁} x) (some {i₂} y)
      with deqt {r = λ _ -> _} C (λ ()) tt i₁ i₂

    deqt (Sig g) f o (some {i}  x) (some y) |  nothing    =    nothing
    deqt (Sig g) f o (some {i}  x) (some y) |  just refl  =    deqt (g i) f o x y
                                                          >>=  just ∘ cong some
    deqt (Fix' F) f o ⟨ x ⟩ ⟨ y ⟩ = deqt F (f ∦ deqt (Fix' F) f) o x y >>= just ∘ cong ⟨_⟩
    deqt {r = r} (Iso C D e) f o x y with e r o

    deqt (Iso C D e) f o x y | ep
      with     _≃_.iso₁ ep x | _≃_.iso₁ ep y |
               deqt C f o (_≃_.from ep x) (_≃_.from ep y)

    deqt (Iso C D e) f o x y | ep | p₁ | p₂ | nothing  = nothing
    deqt (Iso C D e) f o x y | ep | p₁ | p₂ | just p   =
      just (trans (sym p₁) (trans (cong (_≃_.to ep) p) p₂))
    deqtℕ : (m n : Nat2) -> Maybe (m ≡ n)
    deqtℕ = deqt {r = (λ ())} ‵ℕ' (λ ()) tt
    deqtExample₁ : deqtℕ (su ze) (su ze) ≡ just refl
    deqtExample₁ = refl

    deqtExample₂ : deqtℕ (su ze) ze ≡ nothing
    deqtExample₂ = refl
    deqtList :  {AT : Set} ->
                ((x y : AT) -> Maybe (x ≡ y)) -> (l₁ l₂ : (List AT)) -> Maybe (l₁ ≡ l₂)

    deqtList {AT} f x y = deqt {r = const AT} ‵List' (const f) tt x y
    l₁ : (List Nat2)
    l₁ = ze ∷ su ze ∷ []

    l₂ : (List Nat2)
    l₂ = ze ∷ ze ∷ []
    deqtExample₃ : deqtList deqtℕ l₁ l₁ ≡ just refl
    deqtExample₃ = refl

    deqtExample₄ : deqtList deqtℕ l₁ l₂ ≡ nothing
    deqtExample₄ = refl
module zipper where
  module oldZipper where
    open import zipperAux renaming (Nat to Nat2)
    open import appendix

    infixr 7 _∙_
    _∙_ : {A B C : Set} -> (B -> C) -> (A -> B) -> A -> C
    (f ∙ g) x = f (g x)

    --  Awful hack because we have Nat in scope already, and it's mapped
    --  to the built-in nats
    n0 : Nat2
    n0 = ze
    n1 : Nat2
    n1 = su n0
    n2 : Nat2
    n2 = su n1
    n3 : Nat2
    n3 = su n2
    n4 : Nat2
    n4 = su n3
    n5 : Nat2
    n5 = su n4
    n6 : Nat2
    n6 = su n5

    mutual
      Ctx : {I O : Set} -> I ▸ O -> I -> I ▷ O
      Ctx Z'       i r o = ⊥
      Ctx U'       i r o = ⊥
      Ctx (! o′)   i r o = ⊥
      Ctx (I' i′)  i r o = i ≡ i′
      Ctx (F ↗ f ↘ g)  i r o =  ∃ (λ i′ -> ∃ (λ o′ -> f i′ ≡ i  ×  g o ≡ o′  ×  Ctx F i′ (r ∘ f) o′))
      Ctx (F ⊕ G)  i r o = Ctx F i r o ⊎ Ctx G i r o
      Ctx (F ⊗ G)  i r o = Ctx F i r o × ⟦ G ⟧ r o ⊎ ⟦ F ⟧ r o × Ctx G i r o
      Ctx (F ⊚ G)  i r o = ∃ (λ m -> Ctx F m (⟦ G ⟧ r) o × Ctx G i r m)
      Ctx (Sig f)      i r o = ∃ (λ i′ -> Ctx (f i′) i r o)
      Ctx (Iso C D e)  i r o = Ctx C i r o
      Ctx (Fix' F) i r o = ∃ (λ j  ->  Ctx F (inj₁ i) (r ∣ μ F r) j  × Ctxs F j r o)
       
      data  Ctxs {I O : Set} (F : (I ⊎ O) ▸ O) (i : O) (r : Indexed I) : Indexed O where
        empty  :  Ctxs F i r i
        push   :  {j o : O} -> Ctx F (inj₂ i) (r ∣ μ F r) j -> Ctxs F j r o -> Ctxs F i r o
    plug :  {I O : Set} {r : Indexed I} {i : I} {o : O} ->
            (C : I ▸ O) -> Ctx C i r o -> r i -> ⟦ C ⟧ r o
    plug Z'      ()                r
    plug U'      ()                r
    plug (! o)   ()                r
    plug (I' i)  refl              r = r
    plug (F ↗ f ↘ g) (some (some (refl , refl , c))) r = plug F c r
    plug (F ⊕ G) (inj₁   c)          r = inj₁  (plug F  c r)
    plug (F ⊕ G) (inj₂   c)          r = inj₂  (plug G  c r)

    plug (F ⊗ G) (inj₁  (c  , g))  r = plug F c r  ,  g
    plug (F ⊗ G) (inj₂  (f  , c))  r = f           ,  plug G c r

    plug (F ⊚ G) (some (c , d))    r = plug F c (plug G d r)
    plug {r = s} {o = o} (Fix' F) (some {m} (c , cs)) r = unw m cs ⟨ plug F c r ⟩
      where  unw : ∀ m -> Ctxs F m s o -> ⟦ Fix' F ⟧ s m -> ⟦ Fix' F ⟧ s o
             unw  .o  empty            x =  x
             unw  m   (push {o} c cs)  x =  unw o cs ⟨ plug F c x ⟩
    plug (Sig f)    (some {i} c)      r = some (plug (f i) c r)

    plug {r = s} {o = o} (Iso C D e) x r with e s o
    ... | ep = to ep (plug C x r) where open _≃_
    first :  {AT I O : Set} {r : Indexed I} {o : O} -> (C : I ▸ O) ->
             ((i : I) -> r i -> Ctx C i r o -> Maybe AT) -> ⟦ C ⟧ r o -> Maybe AT
    first Z'      k ()
    first U'      k x              = nothing
    first (! o)   k x              = nothing
    first (I' i)  k x              = k i x refl
    first (F ↗ f ↘ g)  k x         = first F  (λ i′  r c -> k (f i′) r (some (some (refl , (refl , c))))) x

    first (F ⊕ G) k (inj₁  x)  = first F  (λ i   r c -> k i r (inj₁  c)) x
    first (F ⊕ G) k (inj₂  x)  = first G  (λ i   r c -> k i r (inj₂  c)) x
    first (F ⊗ G) k (x , y) = plusMaybe  (first F  (λ i r c -> k i r (inj₁  (c  , y  ))) x  )
                                           (first G  (λ i r c -> k i r (inj₂  (x  , c  ))) y  )
    first (F ⊚ G) k x =  first F (λ m s c -> first G (λ i r d -> k i r (some (c , d))) s) x
    first {AT} {I} {O} {r} {o} (Fix' F) k x = fstFix x empty where
       mutual
          fstFix : {m : O} -> μ F r m -> Ctxs F m r o -> Maybe AT
          fstFix ⟨ x ⟩ cs = first F (contFix cs) x
 
          contFix :  {m : O} -> Ctxs F m r o -> (i : I ⊎ O) ->
                     (r ∣ μ F r) i -> Ctx F i (r ∣ μ F r) m -> Maybe AT
          contFix cs  (inj₁  i) r c = k i r (some (c , cs))
          contFix cs  (inj₂  i) r c = fstFix r (push c cs)
    first (Sig f)  k (some {i′} y) = first (f i′) (λ i r c -> k i r (some c)) y

    first {r = r} {o = o} (Iso C D e) k x with e r o 
    ... | ep = first C k (from ep x) where open _≃_
    firstFix : {I O : Set} {r : Indexed I} {o : O} {R : Set} ->
               {m : O} -> (F : (I ⊎ O) ▸ O) ->
               μ F r m -> Ctxs F m r o ->
               ((i : I) -> r i -> Ctx (Fix' F) i r o -> Maybe R) ->
               Maybe R
    firstFix {J} {O} {r} {o} {R} {m} F ⟨ x ⟩ cs k = first F (contFix cs) x
      where
          contFix : {m : O} -> Ctxs F m r o -> (i : J ⊎ O) ->
                    (r ∣ μ F r) i -> Ctx F i (r ∣ μ F r) m -> Maybe R
          contFix cs (inj₁ i) r c = k i r (some (c , cs))
          contFix cs (inj₂ i) r c = firstFix F r (push c cs) k
    next :  {AT I O : Set} {r : Indexed I} {o : O} -> (C : I ▸ O) -> 
            ((i : I) -> r i -> Ctx C i r o -> Maybe AT) -> 
            {i : I} -> Ctx C i r o -> r i -> Maybe AT
    next Z'      k ()                r
    next U'      k ()                r
    next (I' i)  k c                 r = nothing
    next (! o′)  k c                 r = nothing
    next (F ↗ f ↘ g) k (some (some (refl , (p , c)))) r = 
      next F (λ i′ s d -> k (f i′) s (some (some (refl , (p , d))))) c r
    next (F ⊕ G) k (inj₁ c)           r =  next F (λ i s d -> k i s (inj₁ d)) c r
    next (F ⊕ G) k (inj₂ c)           r =  next G (λ i s d -> k i s (inj₂ d)) c r
    next (F ⊗ G) k (inj₁ (c , g))     r =  plusMaybe
                                             (next F (λ i s d -> k i s (inj₁ (d , g))) c r)
                                             (first G (λ i s d -> k i s (inj₂ (plug F c r , d))) g)
    next (F ⊗ G) k (inj₂ (f , c))     r =  next G (λ i s y -> k i s (inj₂ (f , y))) c r
    next (F ⊚ G) k (some (c , d))    r =  plusMaybe
                                             (next G (λ i s d′ -> k i s (some (c , d′))) d r)
                                             (next F (λ m s c′ ->
                                                first G (λ i t d′ -> k i t (some (c′ , d′))) s)
                                                c (plug G d r))
    next (Sig f)  k (some {j} c)       r = next (f j) (λ i s d -> k i s (some d)) c r

    next {R} {J} {O} {s} {o}
         (Fix' F) k (some {i′} (c , cs)) r = next F contFix c r
      where
          contFix : (i : J ⊎ O) -> (s ∣ μ F s) i -> Ctx F i (s ∣ μ F s) i′ -> Maybe R
          contFix (inj₁ i) r c = k i r (some (c , cs))
          contFix (inj₂ i) r c = firstFix F r (push c cs) k

    next (Iso C D e) k x s = next C k x s
    data Loc  {I O : Set} (F : (I ⊎ O) ▸ O) (r : Indexed I) (o : O) : Set where
      loc : {o′ : O} -> ⟦ Fix' F ⟧ r o′ -> Ctxs F o′ r o -> Loc F r o
    Nav : Set₁
    Nav  =  ∀ {I O} {F : (I ⊎ O) ▸ O} {r : Indexed I} {o : O} ->
            Loc F r o -> Maybe (Loc F r o)
    down : Nav
    down {F = F} (loc ⟨ x ⟩ cs) = first F (\  { (inj₁  i) r d -> nothing
                                              ; (inj₂  i) r d -> just (loc r (push d cs)) }) x
    up : Nav
    up          (loc x empty)        =  nothing
    up {F = F}  (loc x (push c cs))  =  just (loc (⟨ plug F c x ⟩) cs)
    right : Nav
    right                      (loc x empty)            = nothing
    right {I} {O} {F} {r} {o}  (loc x (push {m} c cs))  = next F auxf c x
        where  auxf :  (i : I ⊎ O) -> (r ∣ μ F r) i ->
                       Ctx F i (r ∣ μ F r) m -> Maybe (Loc F r o)
               auxf (inj₁  i) r d = nothing
               auxf (inj₂  i) r d = just (loc r (push d cs))
    enter :  ∀ {I O} {F : (I ⊎ O) ▸ O} {r : Indexed I} {o : O} -> 
             ⟦ Fix' F ⟧ r o -> Loc F r o
    enter x = loc x empty

    leave :  ∀ {I O} {F : (I ⊎ O) ▸ O} {r : Indexed I} {o : O} -> 
             Loc F r o -> Maybe (⟦ Fix' F ⟧ r o)
    leave (loc x  empty)       =  just x
    leave (loc x  (push h t))  =  up (loc x (push h t)) >>= leave
    update :  ∀ {I O} {F : (I ⊎ O) ▸ O} {r : Indexed I} ->
              ⟦ Fix' F ⟧ r ⇉ ⟦ Fix' F ⟧ r -> Loc F r ⇉ Loc F r
    update f _ (loc x l) = loc (f _ x) l
    treeB : Rose Nat2 -> Rose Nat2
    treeB t =  fork  n5
                     (fork n4 [] ∷ (fork n3 [] ∷ (fork n2 [] ∷ (fork  n1
                                                                         (t ∷ []) ∷ []))))

    tree : Rose Nat2
    tree = treeB (fork n0 [])
    navTree : Rose Nat2 -> Maybe (Rose Nat2)
    navTree t  =    down (enter (fromRose t))
               >>=  right >>= right >>= right >>= down
               >>=  just ∘ update (const (map ‵Rose′ (const su) tt)) _
               >>=  leave >>= just ∘ toRose
    navTreeExample : navTree tree ≡ just (treeB (fork n1 []))
    navTreeExample = refl
    swapLeft : {n : Nat2} -> Perfect Nat2 {n} -> Maybe (Perfect Nat2 {n})
    swapLeft           p  =    down (enter (fromPerfect p))
                          >>=  just ∘ (update f _)
                          >>=  leave
                          >>=  just ∘ toPerfect

        where  f : (n : Nat2) -> ⟦ `Perfect' ⟧ (T Nat2) n -> ⟦ `Perfect' ⟧ (T Nat2) n
               f n p = fromPerfect (cataPerfect  {R = λ n -> Perfect Nat2 {n}} leaf
                                                 (λ a -> split a ∘ swap) (toPerfect p))
    p₀ : Perfect Nat2
    p₀ = split n0 (leaf , leaf)

    p₁ : Perfect Nat2
    p₁ = split n1 (leaf , leaf)

    p : Perfect Nat2 -> Perfect Nat2 -> Perfect Nat2
    p x y = split n6 ( split n2 (x , y) , split n5 (p₀ , p₁))

    swapLeftExample : swapLeft (p p₀ p₁) ≡ just (p p₁ p₀)
    swapLeftExample = refl
    testZipperAST : Expr Nat2 -> Maybe (Expr Nat2)
    testZipperAST e  =     down (enter (fromExpr e))
                     >>=   just ∘ update (map   `AST' (const su)) _
                     >>=   right >>= just ∘ update (map   `AST' (const su)) _
                     >>=   leave
                     >>=   just ∘ toExpr
module IG where
  data Code : Set₁ where
    U'    :                   Code
    K'    : Set           ->  Code
    R'    : (C : ∞ Code)  ->  Code
    _⊕_   : (C D : Code)  ->  Code
    _⊗_   : (C D : Code)  ->  Code
  data ⟦_⟧ : Code -> Set₁ where
    tt       :                                      ⟦ U'     ⟧
    kk       : {AT : Set}    -> AT              ->  ⟦ K' AT  ⟧
    recc     : {C : ∞ Code}  -> ⟦ ♭ C ⟧         ->  ⟦ R' C   ⟧
    inj₁     : {C D : Code}  -> ⟦ C ⟧           ->  ⟦ C ⊕ D  ⟧
    inj₂     : {C D : Code}  -> ⟦ D ⟧           ->  ⟦ C ⊕ D  ⟧
    _,_  : {C D : Code}  -> ⟦ C ⟧ -> ⟦ D ⟧  ->  ⟦ C ⊗ D  ⟧
  ‵List′ : Set -> Code
  ‵List′ AT = U' ⊕ (K' AT ⊗ R' (♯ ‵List′ AT))
  crush :  {RT : Set}
           (C : Code) -> (RT -> RT -> RT) -> (RT -> RT) -> RT -> ⟦ C ⟧ -> RT

  crush U'       _comb_ frec zero' _            = zero'
  crush (K' _)   _comb_ frec zero' _            = zero'
  crush (R' C)   _comb_ frec zero' (recc x)     = frec (crush (♭ C) _comb_ frec zero' x)
  crush (C ⊕ D)  _comb_ frec zero' (inj₁ x)     = crush C      _comb_ frec zero' x
  crush (C ⊕ D)  _comb_ frec zero' (inj₂ x)     = crush D      _comb_ frec zero' x
  crush (C ⊗ D)  _comb_ frec zero' (x , y)    = (crush C _comb_ frec zero' x) comb (crush D _comb_ frec zero' y)
  size : (C : Code) -> ⟦ C ⟧ -> Nat
  size C = crush C _+_ su ze
  aList : ⟦ ‵List′ ⊤ ⟧
  aList = inj₂ (kk tt , recc (inj₂ (kk tt , recc (inj₁ tt))))

  testSize : size _ aList  ≡ 2
  testSize = refl
module comparison where

  open import Relation.Binary.PropositionalEquality public
  open import Coinduction public
  open import Data.Empty public
  open import Data.Unit public hiding (setoid ; decSetoid ; preorder)
  open import Data.Sum public hiding (map)
  open import Data.Product public hiding (map)
  open import Function public
  open import Level public
  module Regular2PolyP where
    r2pC : Regular.Code -> PolyP.Code
    r2pC Regular.U'         = PolyP.U'
    r2pC Regular.I'         = PolyP.I'
    r2pC (Regular._⊕_ F G)  = (PolyP._⊕_ (r2pC F) (r2pC G))
    r2pC (Regular._⊗_ F G)  = (PolyP._⊗_ (r2pC F) (r2pC G))
    fromRegular : {aT : Set} (C : Regular.Code) -> (Regular.⟦_⟧ C aT) -> (PolyP.⟦_⟧ (r2pC C) ⊥ aT)
    fromRegular Regular.U'          tt           = tt
    fromRegular Regular.I'          x            = x
    fromRegular (Regular._⊕_ F G)   (inj₁  x)    = inj₁  (fromRegular F  x)
    fromRegular (Regular._⊕_ F G)   (inj₂  x)    = inj₂  (fromRegular G  x)
    fromRegular (Regular._⊗_ F G)   (x , y)    = fromRegular F x , fromRegular G y

    fromμRegular : (C : Regular.Code) -> Regular.μ C -> PolyP.μ (r2pC C) ⊥
    fromμRegular C (Regular.⟨_⟩ x) = (PolyP.⟨_⟩ (fromRegular C (Regular.map C (fromμRegular C) x)))
    toRegular : {aT : Set} (C : Regular.Code) -> (PolyP.⟦_⟧ (r2pC C)) ⊥ aT -> (Regular.⟦_⟧ C ) aT
    toRegular Regular.U'          tt          = tt
    toRegular Regular.I'          x           = x
    toRegular (Regular._⊕_ F G)   (inj₁  x)   = inj₁  (toRegular F  x)
    toRegular (Regular._⊕_ F G)   (inj₂  x)   = inj₂  (toRegular G  x)
    toRegular (Regular._⊗_ F G)   (x , y)   = toRegular F x , toRegular G y

    toμRegular : (C : Regular.Code) -> PolyP.μ (r2pC C) ⊥ -> Regular.μ C
    toμRegular C (PolyP.⟨_⟩ x) = (Regular.⟨_⟩ (toRegular C (PolyP.map (r2pC C) id (toμRegular C) x)))
    iso₁R2P :  {R : Set} ->
               (C : Regular.Code) -> (x : (Regular.⟦_⟧ C) R) -> toRegular C (fromRegular C x) ≡ x

    iso₁R2P Regular.U'         _            = refl
    iso₁R2P Regular.I'         _            = refl
    iso₁R2P (Regular._⊕_ F G)  (inj₁  x)    = cong inj₁  (iso₁R2P F  x)
    iso₁R2P (Regular._⊕_ F G)  (inj₂  x)    = cong inj₂  (iso₁R2P G  x)
    iso₁R2P (Regular._⊗_ F G)  (x , y)    = cong₂ _,_ (iso₁R2P F x) (iso₁R2P G y)
    lemma₁ : {R₁ R₂ : Set} (C : Regular.Code)
             {f : R₁ -> R₂} (x : PolyP.⟦_⟧ (r2pC C) ⊥ R₁)
           -> toRegular C (PolyP.map (r2pC C) id f x)
           ≡ Regular.map C f (toRegular C x)
    lemma₁ Regular.U'    _ = refl
    lemma₁ Regular.I'    _ = refl
    lemma₁ (Regular._⊕_ F G) (inj₁ x) = cong inj₁ (lemma₁ F x)
    lemma₁ (Regular._⊕_ F G) (inj₂ x) = cong inj₂ (lemma₁ G x)
    lemma₁ (Regular._⊗_ F G) (x , y)  = cong₂ _,_ (lemma₁ F x) (lemma₁ G y)
    open ≡-Reasoning

    isoμ₁R2P : (C : Regular.Code) (x : Regular.μ C) -> toμRegular C (fromμRegular C x) ≡ x
    isoμ₁R2P C (Regular.⟨_⟩ x) = cong Regular.⟨_⟩ $

        begin

          toRegular C (PolyP.map (r2pC C) id (toμRegular C) (fromRegular C (Regular.map C (fromμRegular C) x)))

        ≡⟨  lemma₁ C _ ⟩

          Regular.map C (toμRegular C) (toRegular C (fromRegular C (Regular.map C (fromμRegular C) x)))
   
        ≡⟨  cong (Regular.map C (toμRegular C)) (iso₁R2P C _) ⟩

          Regular.map C (toμRegular C) (Regular.map C (fromμRegular C) x)

        ≡⟨  Regular.map∘ C ⟩

          Regular.map C (toμRegular C ∘ fromμRegular C) x

        ≡⟨  Regular.map∀ C (isoμ₁R2P C) x ⟩

          Regular.map C id x

        ≡⟨  Regular.mapId C ⟩

          x ∎
    postulate isoμ₂R2P : (C : Regular.Code) (x : PolyP.μ (r2pC C) ⊥) -> fromμRegular C (toμRegular C x) ≡ x

    open CT
    Regular≃PolyP : (C : Regular.Code) -> Regular.μ C ≃ PolyP.μ (r2pC C) ⊥
    Regular≃PolyP C = record  {  from  = fromμRegular   C
                              ;  to    = toμRegular     C
                              ;  iso₁  = isoμ₁R2P       C
                              ;  iso₂  = isoμ₂R2P       C }
  module Regular2Multirec where
    r2mC : Regular.Code -> Multirec.Code ⊤
    r2mC Regular.U'         = Multirec.U'
    r2mC Regular.I'         = Multirec.I' tt
    r2mC (Regular._⊕_ F G)  = (Multirec._⊕_ (r2mC F) (r2mC G))
    r2mC (Regular._⊗_ F G)  = (Multirec._⊗_ (r2mC F) (r2mC G))
    r2m : {aT : Set} -> (C : Regular.Code) -> (Regular.⟦_⟧ C) aT ≡ (Multirec.⟦_⟧ (r2mC C)) (λ _ -> aT) tt
    r2m Regular.U'          = refl
    r2m Regular.I'          = refl
    r2m (Regular._⊕_ F G)   = cong₂ _⊎_ (r2m F) (r2m G)
    r2m (Regular._⊗_ F G)   = cong₂ _×_ (r2m F) (r2m G)
  module Indexed where
   mutual
    Indexed : Set -> Set₁
    Indexed I = I -> Set
  
    _∣_  :  ∀ {I J} -> Indexed I -> Indexed J
        ->  Indexed (I ⊎ J)
    (r ∣ s) (inj₁  i) = r  i
    (r ∣ s) (inj₂  i) = s  i
  
    data Code (I O : Set) : Set₁ where
      U'     :        Code I O
      I'    :  I  ->  Code I O
      !     :  O  ->  Code I O
      _⊕_   :  (F G : Code I O)  -> Code I O
      _⊗_   :  (F G : Code I O)  -> Code I O
      _⊚_   :  {M : Set} -> (F : Code M O)
            ->  (G : Code I M) -> Code I O
      Fix'  :  (F : Code (I ⊎ O) O) -> Code I O
  
    ⟦_⟧  :  {I O : Set} -> Code I O
        ->  Indexed I -> Indexed O
    ⟦ U'      ⟧  r i =  ⊤
    ⟦ I' j   ⟧  r i =  r j
    ⟦ ! j    ⟧  r i =  i ≡ j
    ⟦ F ⊕ G  ⟧  r i =  ⟦ F ⟧ r i  ⊎  ⟦ G ⟧ r i
    ⟦ F ⊗ G  ⟧  r i =  ⟦ F ⟧ r i  ×  ⟦ G ⟧ r i
    ⟦ F ⊚ G  ⟧  r i =  ⟦ F ⟧ (⟦ G ⟧ r) i
    ⟦ Fix' F  ⟧ r i =  μ F r i
  
    data μ {I O : Set} (F : Code (I ⊎ O) O) (r : Indexed I) (o : O) : Set where
      ⟨_⟩ : ⟦ F ⟧ (r ∣ μ F r) o -> μ F r o
  
    _⇉_ : ∀ {I} -> (R S : Indexed I) -> Set
    r ⇉ s = (i : _) -> r i -> s i
  
    _∥_ : ∀ {I J} {A C : Indexed I} {B D : Indexed J} -> A ⇉ C -> B ⇉ D -> (A ∣ B) ⇉ (C ∣ D)
    (f ∥ g) (inj₁  x) z = f  x z
    (f ∥ g) (inj₂  x) z = g  x z
     
    map : {I O : Set} {r s : Indexed I} (F : Code I O) -> (r ⇉ s) -> ⟦ F ⟧ r ⇉ ⟦ F ⟧ s
    map U'        f i _            = tt
    map (I' j)   f i x            = f j x
    map (! j)    f i x            = x
    map (F ⊕ G)  f i (inj₁  x)    = inj₁  (map F  f i x)
    map (F ⊕ G)  f i (inj₂  x)    = inj₂  (map G  f i x)
    map (F ⊗ G)  f i (x , y)      = map F f i x , map G f i y
    map (F ⊚ G)  f i x            = map F (map G f) i x
    map (Fix' F)  f i ⟨ x ⟩       = ⟨ map F (f ∥ map (Fix' F) f) i x ⟩
  
    infix 4 _≅_
    _≅_ : ∀ {I : Set}{r s : Indexed I} (f g : r ⇉ s) -> Set
    f ≅ g = ∀ i x -> f i x ≡ g i x
    
    _⇉∘_ : {Ix : Set} {R S T : Indexed Ix} -> S ⇉ T -> R ⇉ S -> R ⇉ T
    (f ⇉∘ g) ix i = f ix (g ix i)
    
    id⇉ : {Ix : Set} {R : Indexed Ix} -> R ⇉ R
    id⇉ i x = x
    
    ∥∘ : {I J : Set}{r₁ s₁ t₁ : Indexed I}{r₂ s₂ t₂ : Indexed J} -> 
        {f₁ : s₁ ⇉ t₁}{g₁ : r₁ ⇉ s₁}{f₂ : s₂ ⇉ t₂}{g₂ : r₂ ⇉ s₂} ->
        ((f₁ ⇉∘ g₁) ∥ (f₂ ⇉∘ g₂)) ≅ ((f₁ ∥ f₂) ⇉∘ (g₁ ∥ g₂))
    ∥∘ (inj₁ i) x = refl
    ∥∘ (inj₂ j) x = refl
    
    ∥cong : {I J : Set}{r u : Indexed I}{s v : Indexed J}{f₁ f₂ : r ⇉ u}{g₁ g₂ : s ⇉ v} ->
            f₁ ≅ f₂ -> g₁ ≅ g₂ -> (f₁ ∥ g₁) ≅ (f₂ ∥ g₂)
    ∥cong if ig (inj₁ i) x = if i x
    ∥cong if ig (inj₂ j) x = ig j x
{- 
    [,]cong : {I J : Set} {r u : Indexed I} {s v : Indexed J} {f₁ f₂ : r ⇉ u} {g₁ g₂ : s ⇉ v} →
              f₁ ≅ f₂ → g₁ ≅ g₂ → EITHER {C = λ z → (r ∣ s) z → (u ∣ v) z} f₁ g₁ ≅ EITHER f₂ g₂
    [,]cong if ig (inj₁ i) x = if i x
    [,]cong if ig (inj₂ j) x = ig j x

    [,]id : {I J : Set} {r : Indexed (I ⊎ J)} → ((EITHER id⇉ id⇉)) ≅ id⇉ {I ⊎ J} {r}
    [,]id (inj₁ i) x = refl
    [,]id (inj₂ i) x = refl
 -}
    ∥id : {I J : Set} {r : Indexed I} {s : Indexed J}{f : r ⇉ r}{g : s ⇉ s} ->
          f ≅ id⇉ -> g ≅ id⇉ -> (f ∥ g) ≅ id⇉
    ∥id if ig (inj₁ i) x = if i x
    ∥id if ig (inj₂ j) x = ig j x
    
    map-cong : {I O : Set}{r s : Indexed I}{f g : r ⇉ s}->
              (C : Code I O) -> f ≅ g ->
              map C f ≅ map C g
    map-cong U'           ip i x        = refl
    map-cong (I' i′)     ip i x        = ip i′ x
    map-cong (F ⊕ G)     ip i (inj₁ x) = cong inj₁ (map-cong F ip i x)
    map-cong (F ⊕ G)     ip i (inj₂ x) = cong inj₂ (map-cong G ip i x)
    map-cong (F ⊗ G)     ip i (x , y)  = cong₂ _,_ (map-cong F ip i x) (map-cong G ip i y)
    map-cong (F ⊚ G)     ip i x        = map-cong F (map-cong G ip) i x
    map-cong (! o′)      ip i x        = refl
    map-cong (Fix' F)    ip i ⟨ x ⟩    = cong ⟨_⟩ (map-cong F (∥cong ip (map-cong (Fix' F) ip)) i x)
    
    map-id : {I O : Set}{r : Indexed I}(C : Code I O) ->
            map {r = r} C id⇉ ≅ id⇉
    map-id U'          i x        = refl
    map-id (I' i′)    i x        = refl
    map-id (F ⊕ G)    i (inj₁ x) = cong inj₁ (map-id F i x)
    map-id (F ⊕ G)    i (inj₂ x) = cong inj₂ (map-id G i x)
    map-id (F ⊗ G)    i (x , y)  = cong₂ _,_ (map-id F i x) (map-id G i y)
    map-id (F ⊚ G)    i x        = trans (map-cong F (map-id G) i x) (map-id F i x)
    map-id (! o′)     i x        = refl
    map-id (Fix' F)    i ⟨ x ⟩    = cong ⟨_⟩ (trans (map-cong F (∥id (λ _ _ -> refl)
                                                              (map-id (Fix' F))) i x)
                                                  (map-id F i x))
    
    map-∘ : {I O : Set} {r s t : Indexed I}
            (C : Code I O) (f : s ⇉ t) (g : r ⇉ s)
          -> map C (f ⇉∘ g) ≅ map C f ⇉∘ map C g
    map-∘ U'           f g i x        = refl
    map-∘ (I' i′)     f g i x        = refl
    map-∘ (F ⊕ G)     f g i (inj₁ x) = cong inj₁ (map-∘ F f g i x)
    map-∘ (F ⊕ G)     f g i (inj₂ x) = cong inj₂ (map-∘ G f g i x)
    map-∘ (F ⊗ G)     f g i (x , y)  = cong₂ _,_ (map-∘ F f g i x) (map-∘ G f g i y)
    map-∘ (F ⊚ G)     f g i x        = trans (map-cong F (map-∘ G f g) i x)
                                            (map-∘ F (map G f) (map G g) i x)
    map-∘ (! o′)      f g i x        = refl
    map-∘ (Fix' F)     f g i ⟨ x ⟩    = cong ⟨_⟩ (trans (map-cong F (∥cong (λ _ _ -> refl)
                                                                          (map-∘ (Fix' F) f g))
                                                                i x)
                                                (trans (map-cong F ∥∘ i x)
                                                      (map-∘ F (f ∥ map (Fix' F) f)
                                                                (g ∥ map (Fix' F) g) i x)))

    ‵ListC′ : Code (⊤ ⊎ ⊤) ⊤
    ‵ListC′ = U' ⊕ (I' (inj₁ tt) ⊗ I' (inj₂ tt))

    ‵ListF′ : Code ⊤ ⊤
    ‵ListF′ = Fix' ‵ListC′

    data Maybe (A : Set) : Set where
      nothing : Maybe A
      just    : A -> Maybe A

    mapMaybe : {A B : Set} -> (A -> B) -> Maybe A -> Maybe B
    mapMaybe f nothing  = nothing
    mapMaybe f (just x) = just (f x)

    infixl 5 _>>=_
    _>>=_ : {A B : Set} -> Maybe A -> (A -> Maybe B) -> Maybe B
    nothing >>= f = nothing
    just x  >>= f = f x

    plusMaybe : {A : Set} -> Maybe A -> Maybe A -> Maybe A
    plusMaybe (just x) _ = just x
    plusMaybe nothing  y = y

    SemiDec : ∀ {I} -> Indexed I -> Set
    SemiDec r = ∀ i -> (x y : r i) -> Maybe (x ≡ y)

    infixr 5 _∦_
    _∦_ :  {I J : Set} {r : Indexed I} {s : Indexed J} ->
           SemiDec r -> SemiDec s -> SemiDec (r ∣ s)
    (f ∦ g) (inj₁  i)  = f  i
    (f ∦ g) (inj₂  i)  = g  i

    deqt :  {I O : Set} {r : Indexed I} (C : Code I O) ->
            SemiDec r -> SemiDec (⟦ C ⟧ r)
    deqt  U'      f o  tt  tt  = just refl
    deqt  (I' i)  f o  x   y   = f i x y
    deqt (F ⊕ G) f o (inj₁  x) (inj₂  y) =  nothing
    deqt (F ⊕ G) f o (inj₂  x) (inj₁  y) =  nothing

    deqt (F ⊕ G) f o (inj₁  x) (inj₁  y) =  mapMaybe (cong  inj₁)  (deqt F  f o x y)
    deqt (F ⊕ G) f o (inj₂  x) (inj₂  y) =  mapMaybe (cong  inj₂)  (deqt G  f o x y)
    deqt (F ⊗ G) f o (x₁ , x₂) (y₁ , y₂)  =        deqt F  f o x₁  y₁  >>= 
                                              \ l  ->  deqt G  f o x₂  y₂  >>= 
                                              \ r  ->  just (cong₂ _,_ l r)
    deqt (F ⊚ G)  f  o   x     y     = deqt F (deqt G f) o x y 
    deqt (! o)    f  .o  refl  refl  = just refl

    deqt (Fix' F) f o ⟨ x ⟩ ⟨ y ⟩ =      mapMaybe (cong ⟨_⟩) (deqt F (f ∦ deqt (Fix' F) f) o x y)

   cata' :  {I O : Set} {r : Indexed I} {s : Indexed O} -> 
            (C : Code (I ⊎ O) O) -> ⟦ C ⟧ (r ∣ s) ⇉ s -> ⟦ Fix' C ⟧ r ⇉ s
   cata' C φ i ⟨ x ⟩ = φ i (map C (id⇉ ∥ cata' C φ) i x)


  module PolyP2Indexed where
    p2iC : PolyP.Code -> Indexed.Code (⊤ ⊎ ⊤) ⊤

    p2iC ((PolyP._⊚_  F G))  = (Indexed._⊚_ (Indexed.Fix' (p2iC F)) (p2iC G))
    p2iC PolyP.U'            = Indexed.U'
    p2iC PolyP.P'            = Indexed.I' (inj₁ tt)
    p2iC PolyP.I'            = Indexed.I' (inj₂ tt)
    p2iC ((PolyP._⊕_  F G))  = (Indexed._⊕_ (p2iC F) (p2iC G))
    p2iC ((PolyP._⊗_  F G))  = (Indexed._⊗_ (p2iC F) (p2iC G))
    fromP :  {aT rT : Set} (C : PolyP.Code) ->
             (PolyP.⟦_⟧ C aT rT) -> (Indexed.⟦_⟧ (p2iC C) ((Indexed._∣_ (const aT) (const rT))) tt)

    fromP ((PolyP._⊚_ F G))  (PolyP.⟨_⟩ x)  =
      (Indexed.⟨_⟩ (Indexed.map  (p2iC F) ((Indexed._∥_ (λ _ -> fromP G) (λ _ -> fromP ((PolyP._⊚_ F G))))) tt
                                 (fromP F x)))
    fromP PolyP.U'           tt             =  tt
    fromP PolyP.I'           y              =  y
    fromP PolyP.P'           y              =  y
    fromP ((PolyP._⊕_ F G))  (inj₁  x)      =  inj₁ (fromP F x)
    fromP ((PolyP._⊕_ F G))  (inj₂  y)      =  inj₂ (fromP G y)
    fromP ((PolyP._⊗_ F G))  (x , y)        =  fromP F x , fromP G y
    toP :  {aT rT : Set}
           (C : PolyP.Code) -> (Indexed.⟦_⟧ (p2iC C) ((Indexed._∣_ (const aT) (const rT))) tt) -> (PolyP.⟦_⟧ C aT rT)

    toP ((PolyP._⊚_ F G))  (Indexed.⟨_⟩ x)  =
      (PolyP.⟨_⟩ (toP F (Indexed.map (p2iC F) ((Indexed._∥_ (λ _ -> toP G) (λ _ -> toP ((PolyP._⊚_ F G))))) tt x)))
    toP PolyP.U'           tt               = tt
    toP PolyP.I'           x                = x
    toP PolyP.P'           x                = x
    toP ((PolyP._⊕_ F G))  (inj₁  x)        = inj₁ (toP F x)
    toP ((PolyP._⊕_ F G))  (inj₂  y)        = inj₂ (toP G y)
    toP ((PolyP._⊗_ F G))  (x , y)          = toP F x , toP G y

    toμP : {A : Set} (C : PolyP.Code) -> (Indexed.⟦_⟧ (p2iC C) ((Indexed._∣_ (const A) (Indexed.μ (p2iC C) (const A)))) tt) -> PolyP.μ C A
    toμP {A} C x = (PolyP.⟨_⟩ (toP C (Indexed.map (p2iC C) ((Indexed._∥_ (const id) g)) tt x))) where
      g : (i : ⊤) -> Indexed.μ (p2iC C) (const A) i -> _
      g _ (Indexed.⟨_⟩ x) = toμP C x --  unfortunately we cannot inline g above
    fromμP : {aT : Set} (C : PolyP.Code) -> PolyP.μ C aT -> (Indexed.⟦_⟧ (Indexed.Fix' (p2iC C)) (const aT) tt)

    fromμP C (PolyP.⟨_⟩ x) = (Indexed.⟨_⟩ (fromP C (PolyP.map C id (fromμP C) x)))
    open ≡-Reasoning

    isoP₁ : {A R : Set} (C : PolyP.Code) -> (x : (PolyP.⟦_⟧ C A R)) -> toP {A} {R} C (fromP C x) ≡ x

    isoP₁ ((PolyP._⊚_ F G))  (PolyP.⟨_⟩ x)  = cong PolyP.⟨_⟩ $

      begin
    
        (toP F  (Indexed.map (p2iC F) ((Indexed._∥_ (λ _ -> toP     G) (λ _ -> toP    ((PolyP._⊚_ F G))))) tt
                (Indexed.map (p2iC F) ((Indexed._∥_ (λ _ -> fromP   G) (λ _ -> fromP  ((PolyP._⊚_ F G))))) tt (fromP F x))))

      ≡⟨ cong (toP F) (sym (Indexed.map-∘ (p2iC F) _ _ tt (fromP F x))) ⟩

        (toP F (Indexed.map (p2iC F) ((Indexed._⇉∘_  ((Indexed._∥_   (λ _ -> toP    G) (λ _ -> toP    ((PolyP._⊚_ F G)))))
                                                     ((Indexed._∥_   (λ _ -> fromP  G) (λ _ -> fromP  ((PolyP._⊚_ F G))))))) tt (fromP F x)))

      ≡⟨ cong (toP F) (Indexed.map-cong (p2iC F) (λ i x -> sym (Indexed.∥∘ i x)) tt (fromP F x)) ⟩

        (toP F (Indexed.map (p2iC F) ((Indexed._∥_   ((Indexed._⇉∘_  ((λ _ -> toP G))                 ((λ _ -> fromP G))))
                                                     ((Indexed._⇉∘_  ((λ _ -> toP (PolyP._⊚_ F G)))   ((λ _ -> fromP (PolyP._⊚_ F G))))))) tt (fromP F x)))

      ≡⟨ cong (toP F) (Indexed.map-cong (p2iC F) (Indexed.∥cong (λ _ -> isoP₁ G) (λ _ _ -> refl)) tt (fromP F x)) ⟩

        (toP F (Indexed.map (p2iC F) ((Indexed._∥_ Indexed.id⇉ ((Indexed._⇉∘_ (λ _ -> toP ((PolyP._⊚_ F G)))) (λ _ -> fromP ((PolyP._⊚_ F G)))))) tt (fromP F x)))

      ≡⟨ cong (toP F) (Indexed.map-cong (p2iC F) (Indexed.∥cong (λ _ _ -> refl) (λ _ -> isoP₁ (PolyP._⊚_ F G))) tt (fromP F x)) ⟩

        (toP F (Indexed.map (p2iC F) ((Indexed._∥_ Indexed.id⇉ Indexed.id⇉)) tt (fromP F x)))

      ≡⟨ cong (toP F) (Indexed.map-cong (p2iC F) (Indexed.∥id (λ _ _ -> refl) (λ _ _ -> refl)) tt (fromP F x)) ⟩

        (toP F (Indexed.map (p2iC F) Indexed.id⇉ tt (fromP F x)))
    
      ≡⟨ cong (toP F) (Indexed.map-id (p2iC F) tt (fromP F x)) ⟩

        (toP F (fromP F x))

      ≡⟨ isoP₁ F x ⟩

        x ∎
    isoP₁ PolyP.U'           tt             = refl
    isoP₁ PolyP.I'           _              = refl
    isoP₁ PolyP.P'           _              = refl
    isoP₁ ((PolyP._⊕_ F G))  (inj₁  x)      = cong inj₁ (isoP₁ F x)
    isoP₁ ((PolyP._⊕_ F G))  (inj₂  x)      = cong inj₂ (isoP₁ G x)
    isoP₁ ((PolyP._⊗_ F G))  (x , y)        = cong₂ _,_ (isoP₁ F x) (isoP₁ G y)
  module Multirec2Indexed where
    m2iC : {I : Set} -> Multirec.Code I -> Indexed.Code (⊥ ⊎ I) I
    m2iC Multirec.U'            = Indexed.U'
    m2iC (Multirec.I' i)        = Indexed.I' (inj₂ i)
    m2iC (Multirec.! i)         = Indexed.! i
    m2iC ((Multirec._⊕_ F G))   = (Indexed._⊕_ (m2iC F) (m2iC G))
    m2iC ((Multirec._⊗_ F G))   = (Indexed._⊗_ (m2iC F) (m2iC G))
    ↑_ : {I : Set} -> Multirec.Indexed I -> Indexed.Indexed (⊥ ⊎ I)
    (↑ r) (inj₁ ())
    (↑ r) (inj₂ i) = r i
    m2i :  {I : Set} {r : Multirec.Indexed I} {i : I} 
           (C : Multirec.Code I) -> (Multirec.⟦_⟧ C) r i ≡ (Indexed.⟦_⟧ (m2iC C)) (↑ r) i

    m2i Multirec.U'           = refl
    m2i (Multirec.I' i)       = refl
    m2i (Multirec.! i)        = refl
    m2i ((Multirec._⊕_ F G))  = cong₂ _⊎_   (m2i F) (m2i G)
    m2i ((Multirec._⊗_ F G))  = cong₂ _×_   (m2i F) (m2i G)
    ≡1 : {A B : Set} → (A ≡ B) → A → B
    ≡1 refl x = x

    fromM : {Ix : Set} {r : Multirec.Indexed Ix} {i : Ix}
          → (C : Multirec.Code Ix) → Multirec.⟦_⟧ C r i → Indexed.⟦_⟧ (m2iC C) (↑ r) i 
    fromM C x = ≡1 (m2i C) x

    fromμM : {Ix : Set} → (C : Multirec.Code Ix) → (i : Ix)
           → Multirec.μ C i → Indexed.μ (m2iC C) (λ _ → ⊥) i
    fromμM {Ix} C i (Multirec.⟨_⟩ x) = Indexed.⟨_⟩ (Indexed.map (m2iC C) g i (fromM C x)) where
      g : (ix : ⊥ ⊎ Ix) →
          (↑ λ x' → Multirec.μ C x') ix →
          Indexed._∣_ (λ x' → ⊥) (Indexed.μ (m2iC C) (λ x' → ⊥)) ix
      g (inj₁ ()) _
      g (inj₂ j) y = fromμM C j y
    --   g = ⇑ (fromμM C)
  module Indexed2IG where
    i2cC' : {I O : Set} -> Indexed.Code I O -> (I -> IG.Code) -> (O -> IG.Code)
    i2cC' Indexed.U'           r o = IG.U'
    i2cC' (Indexed.I' i)       r o = r i
    i2cC' (Indexed.! i)        r o = IG.K' (o ≡ i)
    i2cC' ((Indexed._⊕_ F G))  r o = (IG._⊕_ (i2cC' F r o) (i2cC' G r o))
    i2cC' ((Indexed._⊗_ F G))  r o = (IG._⊗_ (i2cC' F r o) (i2cC' G r o))
    i2cC' ((Indexed._⊚_ F G))  r o = IG.R' (♯ i2cC' F (i2cC' G r) o)
    i2cC' (Indexed.Fix' F)     r o = IG.R' (♯ i2cC' F ([_,_] r (i2cC' (Indexed.Fix' F) r)) o)

    i2cC : {I O : Set} → Indexed.Code I O → (I → Set) → (O → IG.Code)
    i2cC C r o = i2cC' C (IG.K' ∘ r) o
    fromI2IG :  {I O : Set} {r : I -> IG.Code}
                (C : Indexed.Code I O) (o : O) -> (Indexed.⟦_⟧ C ((IG.⟦_⟧) ∘ r) o) -> (IG.⟦_⟧ (i2cC' C r o))

    fromI2IG Indexed.U'           o tt               = IG.tt
    fromI2IG (Indexed.I' i)       o x                = x
    fromI2IG (Indexed.! i)        o x                = IG.kk x
    fromI2IG ((Indexed._⊕_ F G))  o (inj₁  x)        = IG.inj₁  (fromI2IG F  o x)
    fromI2IG ((Indexed._⊕_ F G))  o (inj₂  x)        = IG.inj₂  (fromI2IG G  o x)
    fromI2IG ((Indexed._⊗_ F G))  o (x , y )       = (IG._,_ (fromI2IG F o x) (fromI2IG G o y))
    fromI2IG ((Indexed._⊚_ F G))  o x                = IG.recc (fromI2IG F o (Indexed.map F (fromI2IG G) o x))
    fromI2IG (Indexed.Fix' F)     o (Indexed.⟨_⟩ x)  = IG.recc (fromI2IG F o (Indexed.map F ([_,_] Indexed.id⇉ (fromI2IG (Indexed.Fix' F))) o x))
     
    fromI2IG' :  {I O : Set} (C : Indexed.Code I O) (r : I → Set) (o : O) →
                 (Indexed.⟦_⟧ C) r o → (IG.⟦_⟧ (i2cC C r o))

    fromI2IG' C r o = fromI2IG {r = IG.K' ∘ r} C o ∘ Indexed.map C (λ _ → IG.kk) o
    toI2IG : {I O : Set} {r : I -> IG.Code} (C : Indexed.Code I O) (o : O) -> IG.⟦_⟧ (i2cC' C r o) -> Indexed.⟦_⟧ C (IG.⟦_⟧ ∘ r) o
    toI2IG Indexed.U'    _ _ = tt
    toI2IG (Indexed.I' _) _ x = x
    toI2IG (Indexed.! _) _ (IG.kk x) = x

    toI2IG (Indexed._⊕_ F G) o (IG.inj₁ x) = inj₁ (toI2IG F o x)
    toI2IG (Indexed._⊕_ F G) o (IG.inj₂ x) = inj₂ (toI2IG G o x)
    toI2IG (Indexed._⊗_ F G) o (IG._,_ x y) = toI2IG F o x , toI2IG G o y

    toI2IG (Indexed._⊚_ F G) o (IG.recc x) =
      Indexed.map F (toI2IG G) o (toI2IG F o x)

    toI2IG (Indexed.Fix' F) o (IG.recc x) =  
      Indexed.⟨_⟩ (Indexed.map F ([_,_] Indexed.id⇉ (toI2IG (Indexed.Fix' F))) o
        (toI2IG F o x))

    open ≡-Reasoning
    iso₁I2IG :  {I O : Set} (C : Indexed.Code I O)
                (r : I -> IG.Code) -> (Indexed._≅_ ((Indexed._⇉∘_ (toI2IG {r = r} C) (fromI2IG C))) (Indexed.id⇉))

    iso₁I2IG ((Indexed._⊚_ F G))  r o x         =

      begin
      
        Indexed.map F (toI2IG G) o (toI2IG F o (fromI2IG F o (Indexed.map F (fromI2IG G) o x)))
    
      ≡⟨  cong (Indexed.map F (toI2IG G) o) (iso₁I2IG F _ o _) ⟩

         Indexed.map F (toI2IG G) o (Indexed.map F (fromI2IG G) o x)

      ≡⟨  sym (Indexed.map-∘ F (toI2IG G) (fromI2IG G) o x) ⟩

        Indexed.map F ((Indexed._⇉∘_ (toI2IG G) (fromI2IG G))) o x

      ≡⟨  Indexed.map-cong F (iso₁I2IG G r) o x ⟩

        Indexed.map F Indexed.id⇉ o x

      ≡⟨  Indexed.map-id F o x ⟩

        x ∎
    iso₁I2IG (Indexed.Fix' F) r o x = iso₁I2IG (Indexed.Fix' F) r o x
    iso₁I2IG  Indexed.U'          r o _           = refl
    iso₁I2IG (Indexed.I' i)       r o _           = refl
    iso₁I2IG (Indexed.!  i)       r o _           = refl
    iso₁I2IG ((Indexed._⊕_ F G))  r o (inj₁   x)  = cong inj₁   (iso₁I2IG F r o   x)
    iso₁I2IG ((Indexed._⊕_ F G))  r o (inj₂   x)  = cong inj₂   (iso₁I2IG G r o   x)
    iso₁I2IG ((Indexed._⊗_ F G))  r o (x , y)   = cong₂ _,_ (iso₁I2IG F r o x) (iso₁I2IG G r o y)
  aListIndexed : (Indexed.⟦_⟧ (Indexed.‵ListF′)) (const ⊤) tt
  aListIndexed = (Indexed.⟨_⟩ (inj₂ (tt , (Indexed.⟨_⟩ (inj₂ (tt , (Indexed.⟨_⟩ (inj₂ (tt , (Indexed.⟨_⟩ (inj₁ tt)))))))))))
  test1 : IG.size (Indexed2IG.i2cC Indexed.‵ListF′ _ _) (Indexed2IG.fromI2IG' Indexed.‵ListF′ _ _ aListIndexed) ≡ 4
  test1 = refl
  test2 : Indexed.Maybe (_ ≡ _)
  test2 =  Indexed.deqt (Indexed.Fix' (Multirec2Indexed.m2iC Multirec.ZigZagC)) (\ _ _ -> \ ()) _ x x
           where x = Multirec2Indexed.fromμM _ _ Multirec.zigZagEnd

  test3 : test2 ≡ Indexed.just refl
  test3 = refl
  test4 : Nat
  test4 =  Indexed.cata' _ f _ (PolyP2Indexed.fromμP _ (Regular2PolyP.fromμRegular _ Regular.aNat))
           where  f : ⊤ -> ⊤ ⊎ Nat -> Nat
                  f _ (inj₁  _)  = 0
                  f _ (inj₂  n)  = 2 + n

  test5 : test4 ≡ 4
  test5 = refl
