{-# LINE 6 "gpid.lhs" #-}
{-#  LANGUAGE ExistentialQuantification  #-}
{-#  LANGUAGE GADTs  #-}
{-#  LANGUAGE KindSignatures  #-}
{-#  LANGUAGE TypeFamilies  #-}
{-#  LANGUAGE TypeOperators  #-}
{-#  LANGUAGE FlexibleContexts  #-}
{-#  LANGUAGE FlexibleInstances  #-}
{-#  LANGUAGE OverlappingInstances  #-}
{-#  LANGUAGE StandaloneDeriving  #-}
{-#  LANGUAGE UndecidableInstances  #-} --  because of X

module ThesisCodeID where

import Data.Typeable

class Constructor c where
  conName   :: t c a -> String
  conFixity :: t c a -> Fixity
  conFixity = const Prefix
  conIsRecord :: t c a -> Bool
  conIsRecord = const False

data Fixity = Prefix | Infix Associativity Int

data Associativity = LeftAssociative | RightAssociative | NotAssociative

data VecNil
data VecCons
data TermLit
data TermIsZero
data TermIf
data TermPair
data EqT_Refl

infixr 6 ::*::


instance Representable (Term a) where
  type Rep (Term a) = RepTerm a
  from (Lit n) = L (CEq_ (Rec_ n))
  from (IsZero n) = R (L (CEq_ (Rec_ n)))
  from (If b x y) = R (R (R (CEq_ (Rec_ b ::*:: (Rec_ x ::*:: Rec_ y)))))
  from (Pair a b) = R (R (L (CEq_ (Rec_ a ::*:: Rec_ b))))

  to (L (CEq_ (Rec_ n))) = Lit n
  to (R (L (CEq_ (Rec_ n)))) = IsZero n
  to (R (R (R (CEq_ (Rec_ b ::*:: Rec_ x ::*:: Rec_ y))))) = If b x y
  to (R (R (L (CEq_ (Rec_ a ::*:: Rec_ b))))) = Pair a b

instance GEq Int where geq = (==)

instance Eq TyEq
instance Eq TyVar
instance Eq Type
instance (Eq a, Eq b) => Eq (a :+: b)

tyVarType = undefined

type List = []
{-# LINE 21 "gpintro.lhs" #-}
infixr 5 :+:
infixr 6 :*:

data aT :+: bT  = L aT | R bT
data aT :*: bT  = aT ::*:: bT
data C cT aT    = C_ aT
data U          = U_
data Var aT     = Var_ aT
data Rec aT     = Rec_ aT
{-# LINE 42 "gpintro.lhs" #-}
class Representable aT where
  type Rep aT
  to    :: Rep aT   -> aT
  from  :: aT       -> Rep aT
{-# LINE 54 "gpintro.lhs" #-}
data List_Nil_
instance Constructor List_Nil_    where conName _ = "[]"

data List_Cons_
instance Constructor List_Cons_   where
  conName _   = ":"
  conFixity _ = Infix RightAssociative 5
 
instance Representable [aT] where
  type Rep [aT]  =    C List_Nil_   U 
                 :+:  C List_Cons_  (Var aT :*: Rec [aT])

  from []                             = L  (C_ U_)
  from (a:as)                         = R  (C_ (Var_ a ::*:: Rec_ as))

  to (L  (C_ U_))                     = []
  to (R  (C_ (Var_ a ::*:: Rec_ as)))  = (a:as)
{-# LINE 89 "gpintro.lhs" #-}
class GEq' aT where
  geq' :: aT -> aT -> Bool
{-# LINE 98 "gpintro.lhs" #-}
instance (GEq' aT, GEq' bT) => GEq' (aT :+: bT) where
  geq' (L  a)  (L  b)  = geq' a b
  geq' (R  a)  (R  b)  = geq' a b
  geq'  _      _       = False
  
instance (GEq' aT, GEq' bT) => GEq' (aT :*: bT) where
  geq' (a ::*:: b) (a' ::*:: b') = geq' a a' && geq' b b'
{-# LINE 112 "gpintro.lhs" #-}
instance GEq' U where
  geq' U_ U_ = True

instance (GEq' aT) => GEq' (C cT aT) where
  geq' (C_ a) (C_ b) = geq' a b
{-# LINE 121 "gpintro.lhs" #-}
instance (GEq aT) => GEq' (Var  aT) where
  geq' (Var_  a)  (Var_  b)  = geq a b

instance (GEq aT) => GEq' (Rec  aT) where
  geq' (Rec_  a)  (Rec_  b)  = geq a b
{-# LINE 131 "gpintro.lhs" #-}
class GEq aT where
  geq :: aT -> aT -> Bool
{-# LINE 137 "gpintro.lhs" #-}
instance GEq Char where
  geq = (==)
{-# LINE 145 "gpintro.lhs" #-}
geqDefault  ::  (Representable aT, GEq' (Rep aT))
            =>  aT -> aT -> Bool
geqDefault x y = geq' (from x) (from y)
{-# LINE 155 "gpintro.lhs" #-}
instance (GEq aT) => GEq [aT] where
  geq = geqDefault
{-# LINE 173 "gpintro.lhs" #-}
class GEnum' aT where
  genum' :: [aT]

instance GEnum' U where
  genum' = [U_]

instance (GEnum   aT) => GEnum' (Rec aT)  where
  genum' = map Rec_     genum

instance (GEnum   aT) => GEnum' (Var aT)  where
  genum' = map Var_     genum

instance (GEnum'  aT) => GEnum' (C cT aT) where
  genum' = map C_       genum'
{-# LINE 195 "gpintro.lhs" #-}
instance (GEnum' aT, GEnum' bT) => GEnum' (aT :+:  bT) where
  genum' = map L genum' ++ map R genum'

instance (GEnum' aT, GEnum' bT) => GEnum' (aT :*:  bT) where
  genum' = [ x ::*:: y | x <- genum', y <- genum' ]
{-# LINE 210 "gpintro.lhs" #-}
genumDefault :: (Representable aT, GEnum' (Rep aT)) => [aT]
genumDefault = map to genum'
{-# LINE 217 "gpintro.lhs" #-}
class GEnum aT where
  genum :: [aT]

instance GEnum Int    where 
  genum = [0..5]
instance GEnum Bool   where
  genum = [True,False]

instance GEnum aT => GEnum [aT] where
  genum = genumDefault
{-# LINE 15 "datatypes.lhs" #-}
data Vec aT nT where
  NilVec   ::                     Vec aT Ze
  ConsVec  :: aT -> Vec aT nT ->  Vec aT (Su nT)
{-# LINE 27 "datatypes.lhs" #-}
data Ze
data Su nT

type T0  = Ze
type T1  = Su  T0
type T2  = Su  T1
{-# LINE 36 "datatypes.lhs" #-}
exampleVec :: Vec Char T2
exampleVec = ConsVec 'p' (ConsVec 'q' NilVec)
{-# LINE 49 "datatypes.lhs" #-}
headVec :: Vec aT (Su nT) -> aT
headVec (ConsVec x _) = x
{-# LINE 57 "datatypes.lhs" #-}
data Term aT where
  Lit     :: Int                              -> Term Int
  IsZero  :: Term Int                         -> Term Bool
  Pair    :: Term aT -> Term bT               -> Term (aT , bT)
  If      :: Term Bool -> Term aT -> Term aT  -> Term aT
{-# LINE 74 "datatypes.lhs" #-}
data Vec2 aT nT   =               nT ~ Ze     => Nil2
                  |  forall  mT.  nT ~ Su mT  => Cons2 aT (Vec2 aT mT)
{-# LINE 100 "datatypes.lhs" #-}
eval :: Term aT -> aT
eval (Lit i)            = i
eval (IsZero t)         = eval t == 0
eval (Pair a b)         = (eval a , eval b)
eval (If p a b)         = if eval p then eval a else eval b
{-# LINE 110 "datatypes.lhs" #-}
deriving instance Show aT => Show (Vec aT nT)
deriving instance Show (Term aT)
{-# LINE 117 "datatypes.lhs" #-}
enumList :: (List aT) -> (List (List aT))
enumList ea = [] : [ x : xs | x <- ea,   xs <- enumList ea ]
{-# LINE 12 "indexingGenerically.lhs" #-}
data aT :=: bT where
  Refl :: aT :=: aT
{-# LINE 25 "indexingGenerically.lhs" #-}
data CEq cT pT qT aT where
  CEq_ :: aT -> CEq cT pT pT aT
{-# LINE 34 "indexingGenerically.lhs" #-}
type C1 cT aT = CEq cT () () aT
{-# LINE 57 "indexingGenerically.lhs" #-}
instance Representable (aT :=: bT) where
  type Rep (aT :=: bT) = CEq EqT_Refl aT bT U
  
  from  Refl        = CEq_ U_
  to    (CEq_ U_)   = Refl
{-# LINE 77 "indexingGenerically.lhs" #-}
instance (GEq' aT) => GEq' (CEq cT pT qT aT) where
  geq' (CEq_ a) (CEq_ b) = geq' a b
 
instance (GEnum' aT) =>   GEnum' (CEq cT pT pT  aT) where
  genum' = map CEq_ genum'

instance                  GEnum' (CEq cT pT qT  aT) where
  genum' = []
{-# LINE 128 "indexingGenerically.lhs" #-}
type family Ex nT
{-# LINE 135 "indexingGenerically.lhs" #-}
instance Representable (Vec aT nT) where
  type Rep (Vec aT nT)  =    CEq  VecNil   nT Ze            U
                        :+:  CEq  VecCons  nT (Su (Ex nT))  (Var aT :*: Rec (Vec aT (Ex nT)))

  from NilVec         = L  (CEq_ U_)
  from (ConsVec h t)  = R  (CEq_ (Var_ h ::*:: Rec_ t))

  to (L  (CEq_ U_))                      = NilVec
  to (R  (CEq_ (Var_ h ::*:: Rec_ t)))   = ConsVec h t
{-# LINE 169 "indexingGenerically.lhs" #-}
type instance Ex (Su mT) = mT
{-# LINE 183 "indexingGenerically.lhs" #-}
data Term3 aT =        aT ~ Int               =>  Lit3      Int
  |                    aT ~ Bool              =>  IsZero3   (Term Int)
  |  forall  bT cT.  aT ~ (bT , cT)   =>  Pair3     (Term bT) (Term cT)
  |                                               If3       (Term Bool) (Term aT) (Term aT)
{-# LINE 192 "indexingGenerically.lhs" #-}
type family Ex1 aT
type family Ex2 aT
{-# LINE 199 "indexingGenerically.lhs" #-}
type family X cT iT aT
{-# LINE 208 "indexingGenerically.lhs" #-}
type RepTerm aT  =    CEq  TermLit     aT  Int
                           (Rec Int)
                           
                 :+:  CEq  TermIsZero  aT  Bool
                           (Rec (Term Int))
                           
                 :+:  CEq  TermPair    aT  (X TermPair T0 aT , X TermPair T1 aT)
                           (    Rec (Term (X TermPair T0 aT))
                           :*:  Rec (Term (X TermPair T1 aT)))
                           
                 :+:  C1   TermIf
                           (Rec (Term Bool) :*: Rec (Term aT) :*: Rec (Term aT))
{-# LINE 225 "indexingGenerically.lhs" #-}
type instance X TermPair T0  (bT , cT) = bT
type instance X TermPair T1  (bT , cT) = cT
{-# LINE 12 "thesisCodeID.hs" #-}
instance (GEq aT) =>  GEq (Vec aT nT)   where
  geq = geqDefault

instance              GEq (Term aT)     where
  geq = geqDefault
{-# LINE 47 "thesisCodeID.hs" #-}
instance (GEnum aT) => GEnum (Vec aT Ze) where
  genum = genumDefault

instance  (GEnum aT, GEnum (Vec aT nT)) =>  GEnum (Vec aT (Su nT)) where
  genum = genumDefault
{-# LINE 67 "thesisCodeID.hs" #-}
instance        GEnum (Term Int)      where
  genum = genumDefault

instance        GEnum (Term Bool)     where
  genum = genumDefault

instance  (GEnum (Term aT), GEnum (Term bT)) =>  GEnum (Term (aT , bT)) where
  genum = genumDefault
{-# LINE 105 "impl.lhs" #-}
data Datatype   = Datatype_ (List TyVar) (List Con)
data Con        = Con_ Constraints (List Type)

data Type   --  abstract
data TyVar  --  abstract

tyVarType :: TyVar -> Type
{-# LINE 122 "impl.lhs" #-}
data Constraints  = (List TyVar) :>: (List TyEq)
data TyEq         = Type :~: Type
{-# LINE 135 "impl.lhs" #-}
findIndices :: Datatype -> (List (TyVar :+: TyVar))
findIndices (Datatype_ vs cs) = [ if v `inArgs` cs then L v else R v | v <- vs ]

inArgs :: TyVar -> [Con] -> Bool
inArgs = undefined
{-# LINE 149 "impl.lhs" #-}
isIndex :: TyVar -> Datatype -> Bool
isIndex t d = R t `elem` findIndices d
{-# LINE 157 "impl.lhs" #-}
findRTs :: (List TyVar) -> (List Con) -> (List Constraints)
findRTs is []                       =  []
findRTs is ((Con_ cts args) : cs)  =  let rs = findRTs is cs 
                                        in  if anyIn is cts then cts :  rs else rs

anyIn :: (List TyVar) -> Constraints -> Bool
anyIn vs (_ :>: teqs) = or [ v `inTEq` teqs | v <- vs ]

inTEq :: TyVar -> (List TyEq) -> Bool
inTEq = undefined
{-# LINE 179 "impl.lhs" #-}
instance Eq Constraints where
  (vs :>: cs) == (ws :>: ds)  =   length vs == length ws &&  cs == subst ws vs ds
 
subst :: (List TyVar) -> (List TyVar) -> (List TyEq) -> (List TyEq)
subst vs ws teqs = undefined
{-# LINE 195 "impl.lhs" #-}
merge :: Constraints -> [Constraints] -> [Constraints]
merge c1                   []                            = [c1]
merge c1@  (vs :>: cs)  (c2@  (ws :>: ds) : css)
  | c1 == c2   = c1  : css
  | otherwise  = c2  : merge c1 css
 
mergeConstraints :: (List Constraints) -> (List Constraints)
mergeConstraints = foldr merge []
{-# LINE 208 "impl.lhs" #-}
rightOnly :: (List (aT :+: bT)) -> (List bT)
rightOnly l = [ x | R x <- l ]
 
allConstraints :: Datatype -> (List Constraints)
allConstraints d@  (Datatype_ _ cons) =  let  is  =  rightOnly (findIndices d)
                                             in mergeConstraints (findRTs is cons)
{-# LINE 223 "impl.lhs" #-}
buildInsts :: Datatype -> (List (List Type))
buildInsts d@  (Datatype_ ts _) = map (instVar ts) cs
        where cs = concat (map (\(_ :>: t) -> t) (allConstraints d))
 
instVar :: (List TyVar) -> TyEq -> (List Type)
instVar []         _               = []
instVar (v : vs)  tEq@  (i :~: t)
  | tyVarType v == i  = t : map tyVarType vs
  | otherwise         = tyVarType v : instVar vs tEq
{-# LINE 8 "futurework.lhs" #-}
data Dynamic = forall  aT. Typeable aT => Dyn aT
{-# LINE 15 "futurework.lhs" #-}
data Term2 aT where
  Const2  :: aT                    -> Term2 aT
  Pair2   :: Term aT -> Term bT    -> Term2 (aT , bT)
  Fst2    :: Term (aT , bT)  -> Term2 aT
  Snd2    :: Term (aT , bT)  -> Term2 bT
