{-# LINE 27 "optimisation.lhs" #-}
{-#  LANGUAGE TypeFamilies  #-}
{-#  LANGUAGE TypeOperators  #-}
{-#  LANGUAGE TypeSynonymInstances  #-}
{-#  LANGUAGE FlexibleContexts  #-}
{-#  LANGUAGE ScopedTypeVariables  #-}

module ThesisCodeOptimisation where

  data Fixity = Prefix | Infix Associativity Int
    deriving (Eq, Show, Ord, Read)

  data Associativity = LeftAssociative | RightAssociative | NotAssociative
    deriving (Eq, Show, Ord, Read)
    
  data U          = U_
  data Var aT     = Var_ aT
  data Rec aT     = Rec_ aT
  data aT :+: bT  = L_ aT | R_ bT
  data aT :*: bT  = aT ::*:: bT
  
  class Representable aT where
    type Rep aT
    to    :: Rep aT   -> aT
    from  :: aT       -> Rep aT

  type List = []
  
  instance Representable (List aT) where
    type Rep (List aT) = U :+: (Var aT :*: Rec (List aT))

    from []                        = L_  U_
    from (h : t)                  = R_  (Var_ h ::*:: Rec_ t)

    to (L_  U_)                     = []
    to (R_  (Var_ h ::*:: Rec_ t))  = h : t
{-# LINE 164 "optimisation.lhs" #-}
  class GEq aT where
    geq :: aT -> aT -> Bool
{-# LINE 172 "optimisation.lhs" #-}
  instance GEq Int where
    geq = (==)
{-# LINE 179 "optimisation.lhs" #-}
  class GEq' aT where
    geq' :: aT -> aT -> Bool
{-# LINE 185 "optimisation.lhs" #-}
  instance GEq' U where
    geq' _ _ = True
    
  instance (GEq' aT, GEq' bT) => GEq' (aT :+: bT) where
    geq'  (L_  x)  (L_  y)  = geq' x y
    geq'  (R_  x)  (R_  y)  = geq' x y
    geq'  _        _        = False

  instance (GEq' aT, GEq' bT) => GEq' (aT :*: bT) where
    geq' (x1 ::*:: y1) (x2 ::*:: y2) = geq' x1 x2 && geq' y1 y2
{-# LINE 202 "optimisation.lhs" #-}
  instance (GEq aT) => GEq' (Rec aT) where
    geq' (Rec_ x) (Rec_ y) = geq x y
    
  instance (GEq aT) => GEq' (Var aT) where
    geq' (Var_ x) (Var_ y) = geq x y
{-# LINE 213 "optimisation.lhs" #-}
  geqDefault :: (Representable aT, GEq' (Rep aT)) => aT -> aT -> Bool
  geqDefault x y = geq' (from x) (from y)
{-# LINE 223 "optimisation.lhs" #-}
  instance (GEq aT) => GEq (List aT) where
    geq = geqDefault
{-# LINE 241 "optimisation.lhs" #-}
  class GEnum' aT where
    genum' :: (List aT)
   
  instance GEnum' U where
    genum' = [U_]

  instance (GEnum aT) => GEnum' (Rec aT) where
    genum' = map Rec_ genum

  instance (GEnum aT) => GEnum' (Var aT) where
    genum' = map Var_ genum
{-# LINE 257 "optimisation.lhs" #-}
  instance (GEnum' aT, GEnum' bT) => GEnum' (aT :+: bT) where
    genum' = map L_ genum' ||| map R_ genum'
   
  infixr 5 |||
  (|||) :: (List aT) -> (List aT) -> (List aT)
{-# LINE 266 "optimisation.lhs" #-}
  []        ||| l = l
  (h : t)  ||| l = h : l ||| t
{-# LINE 275 "optimisation.lhs" #-}
  instance (GEnum' aT, GEnum' bT) => GEnum' (aT :*: bT) where
    genum' = diag (map (\x -> map (\y -> x ::*:: y) genum') genum')
   
  diag :: (List (List aT)) -> (List aT)
{-# LINE 283 "optimisation.lhs" #-}
  diag = concat . foldr skew [] . map (map (: []))

  skew :: (List (List aT)) -> (List (List aT)) -> (List (List aT))
  skew []        l = l
  skew (h : t)  l = h : combine (++) t l

  combine :: (aT -> aT -> aT) -> (List aT) -> (List aT) -> (List aT)
  combine _  xs          []         = xs
  combine _  []         ys          = ys
  combine f  (x : xs)  (y : ys)  = f x y : combine f xs ys
{-# LINE 311 "optimisation.lhs" #-}
  class GEnum aT where
    genum :: (List aT)
   
  genumDefault :: (Representable aT, GEnum' (Rep aT)) => (List aT)
  genumDefault = map to genum'
{-# LINE 329 "optimisation.lhs" #-}
  data Nat = Ze | Su Nat
{-# LINE 333 "optimisation.lhs" #-}
  type RepNat = U :+: Nat
{-# LINE 339 "optimisation.lhs" #-}
  toNat :: RepNat -> Nat
  toNat n =  case n of
               (L_ U_)  -> Ze
               (R_ n)   -> Su n
  
  fromNat :: Nat -> RepNat
  fromNat n =  case n of
                 Ze      -> L_ U_
                 (Su n)  -> R_ n
{-# LINE 357 "optimisation.lhs" #-}
  eqNat :: Nat -> Nat -> Bool
  eqNat m n =  case (m , n) of
                 (Ze      ,  Ze)    -> True
                 (Su m    ,  Su n)  -> eqNat m n
                 (_       ,  _)     -> False
{-# LINE 366 "optimisation.lhs" #-}
  eqU :: U -> U -> Bool
  
  eqU x y =   case (x , y) of
                (U_ , U_) -> True
   
  eqPlus ::  (aT -> aT -> Bool) -> (bT -> bT -> Bool) ->
             aT :+: bT -> aT :+: bT -> Bool
             
  eqPlus ea eb a b =  case (a , b) of
                        (L_   x  ,  L_   y)  -> ea x y
                        (R_   x  ,  R_   y)  -> eb x y
                        (_       ,  _)       -> False
{-# LINE 383 "optimisation.lhs" #-}
  eqRepNat :: RepNat -> RepNat -> Bool
  eqRepNat = eqPlus eqU eqNatFromRep

  eqNatFromRep :: Nat -> Nat -> Bool
  eqNatFromRep m n = eqRepNat (fromNat m) (fromNat n)
{-# LINE 446 "optimisation.lhs" #-}
  enumNat :: (List Nat)
  enumNat = [ Ze ] ||| map Su enumNat
{-# LINE 453 "optimisation.lhs" #-}
  enumU :: (List U)
  enumU = [ U_ ]

  enumPlus :: (List aT) -> (List bT) -> (List (aT :+: bT))
  enumPlus ea eb = map L_ ea ||| map R_ eb
{-# LINE 461 "optimisation.lhs" #-}
  enumRepNat :: (List RepNat)
  enumRepNat = enumPlus enumU enumNatFromRep
{-# LINE 468 "optimisation.lhs" #-}
  enumNatFromRep :: (List Nat)
  enumNatFromRep = map toNat enumRepNat
{-# LINE 606 "optimisation.lhs" #-}
  instance Representable Nat where
    type Rep Nat = U :+: Rec Nat

    to (L_ U_)        = Ze
    to (R_ (Rec_ n))  = Su n
    
    from Ze      = L_ U_
    from (Su n)  = R_ (Rec_ n)
{-# LINE 617 "optimisation.lhs" #-}
  instance GEq Nat where
    geq = geqDefault
{-# LINE 625 "optimisation.lhs" #-}
  gEqNat_geq :: Nat -> Nat -> Bool
  gEqNat_geq = \ (x :: Nat) (y :: Nat) ->  case x of
                                             Ze    ->  case y of
                                                         Ze    -> True
                                                         Su m  -> False
                                             Su m  ->  case y of
                                                         Ze    -> False
                                                         Su n  -> gEqNat_geq m n
{-# LINE 694 "optimisation.lhs" #-}
  instance GEnum Nat where
    genum = map to genum'
{-# LINE 875 "optimisation.lhs" #-}
  data Tree aT = Bin aT (Tree aT) (Tree aT) | Leaf
